import base64
import glob
import json
import logging
import os
import pathlib
import platform
import random
import re
import shlex
import shutil
import socket
import subprocess
import sys
import tempfile
import time
import zipfile
from importlib.metadata import version, PackageNotFoundError
from typing import List, Optional

import requests
import py7zr
from bs4 import BeautifulSoup
from requests.exceptions import HTTPError
from packaging.version import Version

import aniworld.globals as aniworld_globals


def check_dependencies(dependencies: list) -> None:
    resolved_dependencies = []

    for dep in dependencies:
        if dep == "syncplay":
            if platform.system() == "Windows":
                resolved_dependencies.append("SyncplayConsole")
            else:
                resolved_dependencies.append("syncplay")
        else:
            resolved_dependencies.append(dep)

    logging.debug("Checking for %s in path.", resolved_dependencies)
    missing = [dep for dep in resolved_dependencies if shutil.which(dep) is None]

    # TODO: Check if in appdata and return

    if missing:
        download_links = {
            "mpv": "https://mpv.io/installation/",
            "syncplay": "https://syncplay.pl/download/",
            "SyncplayConsole": "https://syncplay.pl/download/",
            "yt-dlp": "https://github.com/yt-dlp/yt-dlp#installation"
        }

        if platform.system() == "Windows" or platform.system() == "Linux":
            logging.debug("Missing dependencies: %s. Attempting to download.", missing)
            missing = [dep.replace("SyncplayConsole", "syncplay") for dep in missing]
            download_dependencies(missing)
        else:
            missing_with_links = [
                f"{dep} (Download: {download_links.get(dep, 'No link available')})"
                for dep in missing
            ]
            logging.critical(
                "Missing dependencies: %s in path. Please install them manually.",
                ', '.join(missing_with_links)
            )
            sys.exit(1)


def fetch_url_content(url: str, proxy: Optional[str] = None, check: bool = True) -> Optional[bytes]:
    if aniworld_globals.DEFAULT_USE_PLAYWRIGHT or os.getenv("USE_PLAYWRIGHT"):
        logging.debug("Now fetching without playwright: %s", url)
        return fetch_url_content_with_playwright(url, proxy, check)

    logging.debug("Now fetching using playwright: %s", url)
    return fetch_url_content_without_playwright(url, proxy, check)


def fetch_url_content_without_playwright(
    url: str, proxy: Optional[str] = None, check: bool = True
) -> Optional[bytes]:
    headers = {
        'User-Agent': aniworld_globals.DEFAULT_USER_AGENT
    }

    logging.debug("Using headers: %s", headers)

    proxies = {}
    if proxy:
        proxies = {
            'http': proxy,
            'https': proxy
        } if proxy.startswith('socks') else {
            'http': f'http://{proxy}',
            'https': f'https://{proxy}'
        }
    elif aniworld_globals.DEFAULT_PROXY:
        default_proxy = aniworld_globals.DEFAULT_PROXY
        proxies = {
            'http': default_proxy,
            'https': default_proxy
        } if default_proxy.startswith('socks') else {
            'http': f'http://{default_proxy}',
            'https': f'https://{default_proxy}'
        }
    else:
        proxies = {
            "http": os.getenv("HTTP_PROXY"),
            "https": os.getenv("HTTPS_PROXY"),
        }

    try:
        response = requests.get(url, headers=headers, proxies=proxies, timeout=300)
        response.raise_for_status()

        if "Deine Anfrage wurde als Spam erkannt." in response.text:
            logging.critical(
                "Your IP address is blacklisted. Please use a VPN, complete the captcha "
                "by opening the browser link, or try again later."
            )

        return response.content

    except requests.exceptions.Timeout as timeout_error:
        logging.critical("Request to %s timed out: %s", url, timeout_error)
        return fetch_url_content_with_playwright(url, proxy, check)

    except requests.exceptions.RequestException as request_error:
        if check:
            logging.critical("Request to %s failed: %s", url, request_error)
        return fetch_url_content_with_playwright(url, proxy, check)


def fetch_url_content_with_playwright(
    url: str, proxy: Optional[str] = None, check: bool = True
) -> Optional[bytes]:

    if "aniworld.to/redirect/" in url:
        return fetch_url_content_without_playwright(url, proxy, check)

    headers = {'User-Agent': aniworld_globals.DEFAULT_USER_AGENT}

    install_and_import("playwright")
    from playwright.sync_api import sync_playwright  # pylint: disable=import-error, import-outside-toplevel

    with sync_playwright() as p:
        options = {'proxy': {'server': proxy}} if proxy else {}
        headless = os.getenv("HEADLESS", not aniworld_globals.IS_DEBUG_MODE)
        browser = p.chromium.launch(headless=headless)

        context = browser.new_context(**options)
        page = context.new_page()
        page.set_extra_http_headers(headers)

        try:
            response = page.goto(url, timeout=10000)
            content = page.content()
            logging.debug(content)

            if page.locator(
                "h1#ddg-l10n-title:has-text('Checking your browser before accessing')"
            ).count() > 0:
                logging.debug("Captcha detected, attempting to solve.")

                for attempt in range(120):
                    page.wait_for_timeout(1000)
                    if page.locator(
                        "h1#ddg-l10n-title:has-text('Checking your browser before accessing')"
                    ).count() == 0:
                        logging.debug("Captcha solved.")
                        break
                    logging.debug("Captcha still present, retry %s/120", attempt + 1)

                if page.locator(
                    "h1#ddg-l10n-title:has-text('Checking your browser before accessing')"
                ).count() > 0:
                    raise TimeoutError("Captcha not solved within the time limit.")

            if response.status != 200:
                raise HTTPError(f"Failed to fetch page: {response.status}")

            page.wait_for_timeout(3000)
            return page.content()

        except (TimeoutError, HTTPError) as error:
            if check:
                logging.critical("Request to %s failed: %s", url, error)
                sys.exit(1)
            return None
        finally:
            context.close()
            browser.close()


def clear_screen() -> None:
    if not aniworld_globals.IS_DEBUG_MODE:
        if platform.system() == "Windows":
            os.system("cls")
        else:
            os.system("clear")


def clean_up_leftovers(directory: str) -> None:
    patterns: List[str] = ['*.part', '*.ytdl', '*.part-Frag*']

    leftover_files: List[str] = []
    for pattern in patterns:
        leftover_files.extend(glob.glob(os.path.join(directory, pattern)))

    for file_path in leftover_files:
        if not os.path.exists(directory):
            logging.warning("Directory %s no longer exists.", directory)
            return

        if os.path.exists(file_path):
            try:
                os.remove(file_path)
                logging.debug("Removed leftover file: %s", file_path)
            except PermissionError:
                logging.warning("Permission denied when trying to remove file: %s", file_path)
            except OSError as e:
                logging.warning("OS error occurred while removing file %s: %s", file_path, e)

    if os.path.exists(directory) and not os.listdir(directory):
        try:
            os.rmdir(directory)
            logging.debug("Removed empty directory: %s", directory)
        except PermissionError:
            logging.warning("Permission denied when trying to remove directory: %s", directory)
        except OSError as e:
            logging.warning("OS error occurred while removing directory %s: %s", directory, e)


def setup_aniskip() -> None:
    script_directory = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if os.name == 'nt':
        mpv_scripts_directory = os.path.join(
            os.environ.get('APPDATA', ''), 'mpv', 'scripts'
        )
    else:
        mpv_scripts_directory = os.path.expanduser('~/.config/mpv/scripts')

    if not os.path.exists(mpv_scripts_directory):
        os.makedirs(mpv_scripts_directory)

    skip_source_path = os.path.join(script_directory, 'aniskip', 'skip.lua')
    skip_destination_path = os.path.join(mpv_scripts_directory, 'skip.lua')

    if os.path.exists(skip_destination_path):
        with open(skip_source_path, 'r', encoding="utf-8") as source_file:
            source_content = source_file.read()

        with open(skip_destination_path, 'r', encoding="utf-8") as destination_file:
            destination_content = destination_file.read()

        if source_content != destination_content:
            logging.debug("Content differs, overwriting skip.lua")
            shutil.copy(skip_source_path, skip_destination_path)
        else:
            logging.debug("skip.lua already exists and is identical, no overwrite needed")
    else:
        logging.debug("Copying skip.lua to %s", mpv_scripts_directory)
        shutil.copy(skip_source_path, skip_destination_path)


def setup_autostart() -> None:
    logging.debug("Copying autostart.lua to mpv script directory")
    script_directory = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if os.name == 'nt':
        mpv_scripts_directory = os.path.join(
            os.environ.get('APPDATA', ''), 'mpv', 'scripts'
        )
    else:
        mpv_scripts_directory = os.path.expanduser('~/.config/mpv/scripts')

    if not os.path.exists(mpv_scripts_directory):
        os.makedirs(mpv_scripts_directory)

    autostart_source_path = os.path.join(script_directory, 'aniskip', 'autostart.lua')
    autostart_destination_path = os.path.join(mpv_scripts_directory, 'autostart.lua')

    if not os.path.exists(autostart_destination_path):
        logging.debug("Copying autostart.lua to %s", mpv_scripts_directory)
        shutil.copy(autostart_source_path, autostart_destination_path)


def setup_autoexit() -> None:
    logging.debug("Copying autoexit.lua to mpv script directory")
    script_directory = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if os.name == 'nt':
        mpv_scripts_directory = os.path.join(
            os.environ.get('APPDATA', ''), 'mpv', 'scripts'
        )
    else:
        mpv_scripts_directory = os.path.expanduser('~/.config/mpv/scripts')

    if not os.path.exists(mpv_scripts_directory):
        os.makedirs(mpv_scripts_directory)

    autoexit_source_path = os.path.join(script_directory, 'aniskip', 'autoexit.lua')
    autoexit_destination_path = os.path.join(mpv_scripts_directory, 'autoexit.lua')

    if not os.path.exists(autoexit_destination_path):
        logging.debug("Copying autoexit.lua to %s", mpv_scripts_directory)
        shutil.copy(autoexit_source_path, autoexit_destination_path)


def get_updated_command_for_mpv(command: List[str], appdata_path: str) -> List[str]:
    command_name = command[0]
    potential_path = os.path.join(appdata_path, command_name)
    if os.path.exists(potential_path):
        command[0] = os.path.join(potential_path, "mpv.exe")
        logging.debug("Updated command for mpv: %s", command)
    return command


def get_updated_command_for_syncplayconsole(command: List[str], appdata_path: str) -> List[str]:
    command_name = command[0]
    potential_path = os.path.join(appdata_path, command_name)
    if os.path.exists(potential_path):
        command[0] = os.path.join(potential_path, "SyncplayConsole.exe")
        logging.debug("Updated command for SyncplayConsole: %s", command)
    else:
        command[0] = os.path.join(appdata_path, "syncplay", "SyncplayConsole.exe")
        logging.debug("Updated command for SyncplayConsole: %s", command)

    for i, arg in enumerate(command):
        if arg == "--player-path" and i + 1 < len(command):
            mpv_path = os.path.join(appdata_path, "mpv", "mpv.exe")
            command[i + 1] = mpv_path
            logging.debug("Updated --player-path argument: %s", command)
    return command


def get_updated_command_for_yt_dlp(command: List[str], appdata_path: str) -> List[str]:
    command_name = command[0]
    potential_path = os.path.join(appdata_path, command_name)
    if os.path.exists(potential_path):
        command[0] = os.path.join(potential_path, "yt-dlp.exe")
        logging.debug("Updated command for yt-dlp: %s", command)
    return command


def execute_command(command: List[str], only_command: bool) -> None:
    logging.debug("Initial command: %s", command)

    if platform.system() == "Windows":
        appdata_path = os.path.join(os.getenv('APPDATA'), 'aniworld')
        logging.debug("AppData path: %s", appdata_path)

        if os.path.exists(appdata_path):
            command_name = command[0]

            if command_name == "mpv":
                command = get_updated_command_for_mpv(command, appdata_path)
            elif command_name == "SyncplayConsole":
                command = get_updated_command_for_syncplayconsole(command, appdata_path)
            elif command_name == "yt-dlp":
                command = get_updated_command_for_yt_dlp(command, appdata_path)

    if only_command:
        command_str = ' '.join(shlex.quote(arg) for arg in command)
        logging.debug("Only command mode: %s", command_str)
        print(command_str)
    else:
        logging.debug("Executing command: %s", command)
        try:
            subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            logging.critical(e)


def raise_runtime_error(message: str) -> None:
    raise RuntimeError(message)


def get_season_episode_count(slug: str, season: str) -> int:
    series_url = f"https://aniworld.to/anime/stream/{slug}/staffel-{season}"

    response = requests.get(series_url, timeout=15)
    soup = BeautifulSoup(response.content, 'html.parser')
    episode_numbers = []
    counter = 1
    while True:
        target = f"{slug}/staffel-{season}/episode-{counter}"

        matching_links = []
        for link in soup.find_all('a', href=True):
            if target in link['href']:
                matching_links.append(link['href'])
        if matching_links:
            episode_numbers.append(counter)
            counter += 1
        else:
            break

    return max(episode_numbers) if episode_numbers else 0


def get_season_episodes(season_url):
    episode_urls = []

    logging.debug("Season URL: %s", season_url)

    parts = season_url.split('/')

    slug = parts[-1]
    season = slug.split('-')[-1]

    slug = parts[-2]

    logging.debug("Slug: %s", slug)
    logging.debug("Season: %s", season)

    for i in range(1, get_season_episode_count(slug, season) + 1):
        episode_urls.append(f"{season_url}/episode-{i}")

    logging.debug("Episode URLs: %s", episode_urls)
    return episode_urls


def get_movies_episode_count(slug: str) -> int:
    movie_url = f"https://aniworld.to/anime/stream/{slug}/filme"
    season_html = fetch_url_content(movie_url)
    if season_html is None:
        return 0
    soup = BeautifulSoup(season_html, 'html.parser')

    movie_numbers = []
    counter = 1
    while True:
        target = f"{slug}/filme/film-{counter}"

        matching_links = []
        for link in soup.find_all('a', href=True):
            if target in link['href']:
                matching_links.append(link['href'])
        if matching_links:
            movie_numbers.append(counter)
            counter += 1
        else:
            break

    return max(movie_numbers) if movie_numbers else 0


def get_season_data(anime_slug: str):
    base_url_template = "https://aniworld.to/anime/stream/{anime}/"
    base_url = base_url_template.format(anime=anime_slug)

    logging.debug("Fetching Base URL %s", base_url)
    main_html = fetch_url_content(base_url)
    if main_html is None:
        logging.critical("Failed to retrieve main page.")
        sys.exit(1)

    soup = BeautifulSoup(main_html, 'html.parser')
    season_meta = soup.find('meta', itemprop='numberOfSeasons')
    number_of_seasons = int(season_meta['content']) if season_meta else 0

    movies = False

    if soup.find('a', title='Alle Filme'):
        number_of_seasons -= 1
        movies = True

    season_data = []
    for i in range(1, number_of_seasons + 1):
        season_url = f"{base_url}staffel-{i}"
        season_data.extend(get_season_episodes(season_url))

    if movies:
        movie_data = []
        number_of_movies = get_movies_episode_count(anime_slug)
        for i in range(1, number_of_movies + 1):
            movie_data.append(f"https://aniworld.to/anime/stream/{anime_slug}/filme/film-{i}")

        season_data.extend(movie_data)

    return season_data


def set_terminal_size(columns: int = None, lines: int = None):
    logging.debug("Setting terminal size to %s columns and %s lines.", columns, lines)
    system_name = platform.system()

    if not columns or not lines:
        columns, lines = aniworld_globals.DEFAULT_TERMINAL_SIZE

    if system_name == 'Darwin':
        os.system(f"printf '\033[8;{lines};{columns}t'")

    # TODO: Windows and Linux support


def get_season_and_episode_numbers(episode_url: str) -> tuple:
    logging.debug("Extracting season and episode numbers from URL: %s", episode_url)
    if "staffel" in episode_url and "episode" in episode_url:
        matches = re.findall(r'\d+', episode_url)
        season_episode = int(matches[-2]), int(matches[-1])
    elif "filme" in episode_url:
        movie_number = re.findall(r'film-(\d+)', episode_url)
        season_episode = 0, int(movie_number[0]) if movie_number else 1
    else:
        logging.error("URL format not recognized: %s", episode_url)
        raise ValueError("URL format not recognized")
    logging.debug("Extracted season and episode numbers: %s", season_episode)
    return season_episode


def ftoi(value: float) -> str:
    return str(int(value * 1000))


def get_version():
    try:
        __version__ = version("aniworld")
    except PackageNotFoundError:
        __version__ = "0.0.0"

    return f" v.{__version__}"


def get_latest_github_version():
    repo = "phoenixthrush/aniworld-downloader"
    api_url = f"https://api.github.com/repos/{repo}/releases/latest"

    try:
        response_content = fetch_url_content(api_url, check=False)
        if not response_content:
            logging.error("Failed to fetch latest release from %s", repo)
            return ""

        release_data = json.loads(response_content)
        latest_version = release_data.get('tag_name', '')
        logging.debug("Latest GitHub version: %s", latest_version)
        return latest_version
    except json.JSONDecodeError as e:
        logging.error("Error decoding JSON response from %s: %s", repo, e)
    except requests.exceptions.RequestException as e:
        logging.error("Unexpected error fetching latest release from %s: %s", repo, e)
    return ""


def is_version_outdated():
    current_version = get_version()
    latest_version = get_latest_github_version()

    if not current_version or not latest_version:
        logging.error("Could not determine version information.")
        return False

    current_version = Version(current_version.strip().lstrip('v').lstrip('.'))
    latest_version = Version(latest_version.strip().lstrip('v').lstrip('.'))

    logging.debug("Current version: %s, Latest version: %s", current_version, latest_version)

    return current_version < latest_version


def get_language_code(language: str) -> str:
    logging.debug("Getting language code for: %s", language)
    return {
        "German Dub": "1",
        "English Sub": "2",
        "German Sub": "3"
    }.get(language, "")


def get_language_string(lang_key: int) -> str:
    lang_map = {
        1: "German Dub",
        2: "English Sub",
        3: "German Sub"
    }
    return lang_map.get(lang_key, "Unknown Language")


def get_github_release(repo: str) -> dict:
    api_url = f"https://api.github.com/repos/{repo}/releases/latest"

    try:
        response_content = fetch_url_content(api_url, check=False)
        if not response_content:
            logging.error("Failed to fetch latest release from %s", repo)
            return {}

        release_data = json.loads(response_content)
        return {
            asset['name']: asset['browser_download_url']
            for asset in release_data.get('assets', [])
        }
    except json.JSONDecodeError as e:
        logging.error("Error decoding JSON response from %s: %s", repo, e)
    except requests.exceptions.RequestException as e:
        logging.error("Unexpected error fetching latest release from %s: %s", repo, e)
    return {}


def download_dependencies(dependencies: list):
    logging.debug("Dependencies to download: %s", dependencies)

    if platform.system() == "Linux" or platform.system() == "Darwin":
        logging.debug("Installing using Package-Manager...!")
        install_packages(get_package_manager(), dependencies)
        return

    if platform.system() != "Windows":
        logging.debug("Not on Windows, skipping dependency download.")
        return

    dependencies = [dep for dep in dependencies if not shutil.which(dep)]
    if not dependencies:
        logging.debug("All required dependencies are already in PATH. No downloads needed.")
        return

    appdata_path = os.path.join(os.getenv('APPDATA'), 'aniworld')
    logging.debug("Creating appdata path: %s", appdata_path)
    os.makedirs(appdata_path, exist_ok=True)

    for dep in dependencies:
        dep_path = os.path.join(appdata_path, dep)
        if os.path.exists(dep_path):
            logging.debug("%s already exists. Skipping download.", dep_path)
            continue

        logging.debug("Creating directory for %s at %s", dep, dep_path)
        os.makedirs(dep_path, exist_ok=True)
        download_and_extract_dependency(dep, dep_path, appdata_path)

    logging.debug("Windows dependencies downloaded.")


def download_and_extract_dependency(dep: str, dep_path: str, appdata_path: str):
    if dep == 'mpv':
        if platform.system() == "Windows":
            logging.debug("Downloading mpv...")
            print_progress_info("Downloading mpv...")
        else:
            logging.info("Downloading mpv...")
        download_mpv(dep_path, appdata_path)
    elif dep == 'syncplay':
        if platform.system() == "Windows":
            logging.debug("Downloading Syncplay...")
            print_progress_info("Downloading Syncplay...")
        else:
            logging.info("Downloading Syncplay...")
        download_syncplay(dep_path)
    elif dep == 'yt-dlp':
        if platform.system() == "Windows":
            logging.debug("Downloading yt-dlp...")
            print_progress_info("Downloading yt-dlp...")
        else:
            logging.info("Downloading yt-dlp...")
        download_yt_dlp(dep_path)


def download_mpv(dep_path: str, appdata_path: str):
    direct_links = get_github_release("shinchiro/mpv-winbuild-cmake")
    try:
        logging.debug("Checking for AVX2 support...")
        avx2_supported = check_avx2_support()
        if avx2_supported:
            logging.debug("AVX2 is supported.")
        else:
            logging.debug("AVX2 is not supported.")
    except Exception:  # pylint: disable=broad-exception-caught  # TODO explicitly specify
        logging.debug("Exception while checking for avx2, defaulting support to False.")
        avx2_supported = False
    pattern = r'mpv-x86_64-\d{8}-git-[a-f0-9]{7}\.7z'
    if avx2_supported:
        logging.debug("AVX2 is supported, using mpv v3.")
        pattern = r'mpv-x86_64-v3-\d{8}-git-[a-f0-9]{7}\.7z'
    else:
        logging.debug("AVX2 is not supported.")

    logging.debug("Downloading %s", pattern)

    direct_link = next(
        (link for name, link in direct_links.items()
         if re.match(pattern, name)),
        None
    )

    logging.debug("Direct link: %s", direct_link)

    if not direct_link:
        logging.error("No download link found for MPV. Please download it manually.")
        return
    logging.debug(direct_link)

    zip_path = os.path.join(appdata_path, 'mpv.7z')
    logging.debug("Downloading MPV from %s to %s", direct_link, zip_path)
    url_content = fetch_url_content(direct_link)

    logging.debug("Saving url content.")
    with open(zip_path, 'wb') as f:
        f.write(url_content)

    logging.debug("Unpacking %s to %s", zip_path, dep_path)
    with py7zr.SevenZipFile(zip_path, mode='r') as archive:
        archive.extractall(path=dep_path)

    logging.debug("Removing %s after unpacking", zip_path)
    os.remove(zip_path)


def download_syncplay(dep_path: str):
    logging.debug("Getting latest syncplay direct link.")
    direct_links = get_github_release("Syncplay/syncplay")
    direct_link = next(
        (link for name, link in direct_links.items()
         if re.match(r'Syncplay_\d+\.\d+\.\d+_Portable\.zip', name)),
        None
    )
    if not direct_link:
        logging.error("No download link found for Syncplay. Please install it manually.")
        return
    logging.debug(direct_link)

    exe_path = os.path.join(dep_path, 'syncplay.zip')
    logging.debug("Downloading Syncplay from %s to %s", direct_link, exe_path)
    url_content = fetch_url_content(direct_link)
    with open(exe_path, 'wb') as f:
        f.write(url_content)

    logging.debug("Unpacking %s to %s", exe_path, dep_path)
    shutil.unpack_archive(exe_path, dep_path)

    logging.debug("Removing %s after unpacking", exe_path)
    os.remove(exe_path)


def download_yt_dlp(dep_path: str):
    url = "https://github.com/yt-dlp/yt-dlp/releases/latest/download/yt-dlp.exe"
    exe_path = os.path.join(dep_path, 'yt-dlp.exe')

    logging.debug("Downloading yt-dlp from %s to %s", url, exe_path)
    url_content = fetch_url_content(url)

    logging.debug("Saving url content.")
    with open(exe_path, 'wb') as f:
        f.write(url_content)


def is_tail_running():
    try:
        result = subprocess.run(
            ["sh", "-c", "ps aux | grep 'tail -f.*/aniworld.log' | grep -v grep"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=False  # TODO fix on MacOS
        )
        return result.returncode == 0
    except subprocess.CalledProcessError as e:
        logging.error("CalledProcessError checking if tail is running: %s", e)
        return False
    except subprocess.SubprocessError as e:
        logging.error("SubprocessError checking if tail is running: %s", e)
        return False


def check_avx2_support() -> bool:
    if platform.system() != "Windows":
        logging.debug("AVX2 check is only supported on Windows, defaulting to False.")
        return False

    try:
        if shutil.which("wmic"):
            logging.debug("wmic is in path.")
            cpu_info = subprocess.run(
                ['wmic', 'cpu', 'get',
                 'Caption, Architecture, DataWidth, Manufacturer, ProcessorType, Status'],
                capture_output=True, text=True, check=False
            )
            logging.debug("CPU Info: %s", cpu_info.stdout.decode('utf-8', errors='replace'))
            if 'avx2' in cpu_info.stdout.decode('utf-8', errors='replace').lower():
                logging.debug("AVX2 is supported.")
                return True
        else:
            logging.debug("wmic is not in path, defaulting to False.")
        return False
    except subprocess.CalledProcessError as e:
        logging.error("Error checking AVX2 support, defaulting to False.: %s", e)
        return False
    except subprocess.SubprocessError as e:
        logging.error("Subprocess error checking AVX2 support, defaulting to False.: %s", e)
        return False


def remove_path(path):
    try:
        if os.path.isfile(path):
            logging.debug("Removing file: %s", path)
            os.remove(path)
        elif os.path.isdir(path):
            logging.debug("Removing directory: %s", path)
            shutil.rmtree(path)
        logging.debug("Removed %s", path)
    except OSError as e:
        logging.error("Error removing %s: %s", path, e)


def get_mpv_directory():
    if platform.system() == "Windows":
        return os.path.join(os.getenv('APPDATA'), 'mpv')

    return os.path.join(os.getenv('HOME'), '.config', 'mpv')


def get_aniworld_data_directory():
    if platform.system() == "Windows":
        return os.path.join(os.getenv('APPDATA'), 'aniworld', 'anime4k')

    return os.path.join(os.getenv('HOME'), '.aniworld', 'anime4k')


def get_anime4k_download_link(mode: str):
    os_type = "Windows" if platform.system() == "Windows" else "Mac_Linux"
    return (
        f"https://github.com/Tama47/Anime4K/releases/download/v4.0.1/"
        f"GLSL_{os_type}_{mode}-end.zip"
    )


def download_anime4k(mode: str):
    download_link = get_anime4k_download_link(mode)
    logging.debug("Downloading Anime4k from %s", download_link)

    anime4k_path = get_aniworld_data_directory()
    shaders_path = os.path.join(anime4k_path, f"GLSL_{platform.system()}_{mode}-end")
    archive_path = os.path.join(shaders_path, "anime4k.zip")

    os.makedirs(shaders_path, exist_ok=True)
    logging.debug("Created shaders path: %s", shaders_path)

    content = fetch_url_content(download_link)
    logging.debug("Fetched content from %s", download_link)

    with open(archive_path, 'wb') as f:
        f.write(content)
    logging.debug("Saved archive to %s", archive_path)

    logging.debug("Extracting package")
    with zipfile.ZipFile(archive_path, 'r') as zip_ref:
        zip_ref.extractall(shaders_path)
    logging.debug("Extracted package to %s", shaders_path)

    remove_path(archive_path)
    logging.debug("Removed archive: %s", archive_path)
    remove_path(os.path.join(shaders_path, "__MACOSX"))
    logging.debug("Removed __MACOSX directory if it existed.")


def remove_anime4k_files():
    mpv_directory = get_mpv_directory()
    input_conf_path = os.path.join(mpv_directory, "input.conf")

    logging.debug("Removing existing configuration files.")
    remove_path(input_conf_path)
    remove_path(os.path.join(mpv_directory, "mpv.conf"))
    remove_path(os.path.join(mpv_directory, "shaders"))


def set_anime4k_config(mode: str):
    anime4k_path = get_aniworld_data_directory()
    shaders_path = os.path.join(anime4k_path, f"GLSL_{platform.system()}_{mode}-end")
    mpv_directory = get_mpv_directory()

    os.makedirs(mpv_directory, exist_ok=True)
    logging.debug("Created MPV directory: %s", mpv_directory)

    input_conf_path = os.path.join(mpv_directory, "input.conf")
    if os.path.exists(input_conf_path):
        logging.debug("Found existing input.conf at %s", input_conf_path)
        with open(input_conf_path, "r", encoding='utf-8') as file:
            lines = file.readlines()

        current_mode = None
        for line in lines:
            if "lower-end" in line:
                current_mode = "Low"
                break
            if "higher-end" in line:
                current_mode = "High"
                break

        if current_mode == mode:
            msg = f"Anime4K: Current mode is already set to {mode}. No changes made."
            print(msg)
            logging.debug(msg)
            sys.exit()

        remove_anime4k_files()

    logging.debug("Copying shaders from %s to %s", shaders_path, mpv_directory)
    for item in os.listdir(shaders_path):
        source = os.path.join(shaders_path, item)
        destination = os.path.join(mpv_directory, item)
        if os.path.isdir(source):
            shutil.copytree(source, destination, dirs_exist_ok=True)
            logging.debug("Copied directory %s to %s", source, destination)
        else:
            shutil.copy(source, destination)
            logging.debug("Copied file %s to %s", source, destination)


def setup_anime4k(mode: str):
    if mode == "Remove":
        remove_anime4k_files()
        print("Anime4K: Uninstalled.")
        sys.exit()

    download_anime4k(mode)
    set_anime4k_config(mode)

    print(f"Anime4K: Installed using mode: {mode}.")
    sys.exit()


def process_episode_file_line(line: str) -> tuple:
    if "https://aniworld.to/anime/stream/" in line:
        if "/staffel-" in line and "/episode-" in line:
            slug = line.split('/')[5]
            return [line], slug
        if "/staffel-" in line:
            return get_season_episodes(line), line.split('/')[-2]

        slug = line.split('/')[-1]
        return list(get_season_data(slug)), slug

    return [], None


def read_episode_file(file: str) -> dict:
    animes = {}

    try:
        with open(file, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue

                episode, slug = process_episode_file_line(line)
                if episode:
                    if slug not in animes:
                        animes[slug] = []
                    animes[slug].extend(episode)
    except FileNotFoundError:
        msg = "The specified episode file was not found!"
        logging.debug(msg)
        print(msg)

    return animes


def check_package_installation(package_name: str = "aniworld"):
    site_packages = next(p for p in sys.path if 'site-packages' in p)

    package_path = pathlib.Path(site_packages) / package_name
    git_path = package_path / "../../../../.git"

    if git_path.exists():
        return "clone"

    dist_info_path = pathlib.Path(site_packages) / f"{package_name}-*.dist-info"
    dist_info_dirs = list(dist_info_path.parent.glob(f"{package_name}-*.dist-info"))

    if dist_info_dirs:
        direct_url_file = dist_info_dirs[0] / "direct_url.json"
        if direct_url_file.exists():
            return "git"

    return "pypi"


def remove_files(paths):
    for path in paths:
        try:
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)
        except OSError as e:
            print(f"Error removing {path}: {e}")


def get_uninstall_paths():
    base_path = (
        os.getenv('APPDATA') if platform.system() == "Windows"
        else os.path.expanduser("~/.config/mpv")
    )

    config_dir = (
        os.path.expanduser("~/.aniworld") if platform.system() != "Windows"
        else os.path.join(os.getenv('APPDATA'), "aniworld")
    )

    return [
        os.path.join(base_path, "scripts", "autoexit.lua"),
        os.path.join(base_path, "scripts", "autostart.lua"),
        os.path.join(base_path, "scripts", "skip.lua"),
        os.path.join(base_path, "input.conf"),
        os.path.join(base_path, "mpv.conf"),
        os.path.join(base_path, "shaders"),
        config_dir,
    ]


def execute_detached_command_windows(command):
    subprocess.Popen(  # pylint: disable=consider-using-with
        f'timeout 3 >nul & {" ".join(command)}',
        shell=True,
        creationflags=subprocess.CREATE_NEW_CONSOLE
    )


def uninstall_aniworld():
    if shutil.which("pip"):
        command = ["pip", "uninstall", "aniworld", "-y"]

        if platform.system() == "Windows":
            execute_detached_command_windows(command)
        else:
            execute_command(command, only_command=False)


def self_uninstall():
    paths = get_uninstall_paths()

    logging.debug("Removed Files:\n%s", paths)
    remove_files(paths)

    logging.debug("Uninstalling using pip.")
    print("Uninstalling, please wait up to 3 seconds...")
    uninstall_aniworld()
    sys.exit()


def get_component_paths():
    base_path = (
        os.path.join(os.getenv('APPDATA'), "aniworld")
        if platform.system() == "Windows"
        else os.path.expanduser("~/.aniworld")
    )

    return {
        "mpv": os.path.join(base_path, "mpv"),
        "yt-dlp": os.path.join(base_path, "yt-dlp"),
        "syncplay": os.path.join(base_path, "syncplay")
    }


def update_component(component: str):
    paths = get_component_paths()

    components = ["mpv", "yt-dlp", "syncplay"] if component == "all" else [component]

    for comp in components:
        remove_path(paths[comp])
        logging.debug("Removed: %s", comp)
        logging.debug("Downloading component: %s", comp)
        download_dependencies([comp])
        if not platform.system() == "Darwin":
            print(f"Installed latest {comp} version.")


def print_progress_info(msg: str):
    command = f"""cmd /c echo {msg.replace('"', "'")} """
    execute_command(command, only_command=False)


def get_anime_season_title(slug: str, season: int) -> str:
    # TODO this should be replaced with a logic for the actual name for each season using api
    # this will also be called for each season but for now once
    logging.debug("Fetching %s season %s name", slug, season)

    season_html = fetch_url_content(f"https://aniworld.to/anime/stream/{slug}/staffel-{season}")

    if not season_html:
        logging.error("Failed to fetch content for %s season %s", slug, season)
        return slug.replace("-", " ").title()

    soup = BeautifulSoup(season_html, 'html.parser')

    series_div = soup.find('div', class_='series-title')

    if series_div:
        name = series_div.find('h1').find('span').text
    else:
        logging.warning("No series title found for %s season %s, using slug instead", slug, season)
        name = slug.replace("-", " ").title()

    logging.debug("Anime season title: %s", name)

    return name


def countdown():
    try:
        msg = "You now have 3 seconds to press CTRL+C to exit!"
        if not platform.system() == "Windows":
            print(msg)
        else:
            print_progress_info(msg)
        time.sleep(3)
    except KeyboardInterrupt:
        print("\nExiting...")
        sys.exit()


def sanitize_path(path):
    invalid_chars = r'\/:*?"<>|&'

    translation_table = str.maketrans('', '', invalid_chars)

    sanitized_path = path.translate(translation_table)

    return sanitized_path


def get_package_manager():
    try:
        system = platform.system()
        package_manager = 'unknown'

        if system == "Darwin":
            package_manager = 'brew' if shutil.which("brew") else 'unknown'

        if os.path.exists('/etc/os-release'):
            with open('/etc/os-release', encoding='utf-8') as f:
                os_release_info = f.read().lower()

            if 'arch' in os_release_info:
                package_manager = 'pacman'
            if 'ubuntu' in os_release_info or 'debian' in os_release_info:
                package_manager = 'apt'
            if 'fedora' in os_release_info:
                package_manager = 'dnf'
            if 'centos' in os_release_info or 'rhel' in os_release_info:
                package_manager = 'yum'
            if 'gentoo' in os_release_info:
                package_manager = 'emerge'
            if 'opensuse' in os_release_info:
                package_manager = 'zypper'
            if 'alpine' in os_release_info:
                package_manager = 'apk'

        return package_manager

    except FileNotFoundError as e:
        return f'Error: {e}'
    except PermissionError as e:
        return f'Error: {e}'


def install_packages(package_manager, packages):
    try:
        if package_manager == 'pacman':
            subprocess.run(
                ['pkexec', 'pacman', '-S', '--noconfirm'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'apt':
            subprocess.run(
                ['pkexec', 'apt-get', 'install', '-y'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'dnf':
            subprocess.run(
                ['pkexec', 'dnf', 'install', '-y'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'yum':
            subprocess.run(
                ['pkexec', 'yum', 'install', '-y'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'emerge':
            subprocess.run(
                ['pkexec', 'emerge'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'zypper':
            subprocess.run(
                ['pkexec', 'zypper', 'install', '-y'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'apk':
            subprocess.run(
                ['pkexec', 'apk', 'add'] + packages,
                stdout=subprocess.DEVNULL,
                check=False
            )
        elif package_manager == 'brew':
            msg = (
                f'Please update "{packages[0]}" manually as it is not currently '
                'supported yet on MacOS!'
            )
            logging.debug(msg)
            print(msg)
        else:
            print(f'Package manager "{package_manager}" not supported or unknown.')
    except subprocess.SubprocessError as e:
        print(f'Error while installing: {e}')


def open_terminal_with_command(command):
    terminal_emulators = [
        ('gnome-terminal', ['gnome-terminal', '--', 'bash', '-c', f'{command}; exec bash']),
        ('xterm', ['xterm', '-hold', '-e', command]),
        ('konsole', ['konsole', '--hold', '-e', command])
    ]

    for terminal, cmd in terminal_emulators:
        try:
            subprocess.Popen(cmd)  # pylint: disable=consider-using-with
            return
        except FileNotFoundError:
            logging.debug("%s not found, trying the next option.", terminal)
        except subprocess.SubprocessError as e:
            logging.error("Error opening terminal with %s: %e", terminal, e)

    logging.error(
        "No supported terminal emulator found. "
        "Please install gnome-terminal, xterm, or konsole."
    )


def get_random_anime(genre: str) -> str:
    url = 'https://aniworld.to/ajax/randomGeneratorSeries'
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'Accept': '*/*',
        'Sec-Fetch-Site': 'same-origin',
        'Accept-Language': 'en-GB,en;q=0.9',
        'Accept-Encoding': 'gzip, deflate, br',
        'Sec-Fetch-Mode': 'cors',
        'Origin': 'https://aniworld.to',
        'User-Agent': aniworld_globals.DEFAULT_USER_AGENT,
        'Referer': 'https://aniworld.to/random',
        'X-Requested-With': 'XMLHttpRequest',
    }
    data = {
        'productionStart': 'all',
        'productionEnd': 'all',
        'genres[]': genre
    }

    response = requests.post(url, headers=headers, data=data, timeout=15)
    logging.debug("Response Status Code: %s", response.status_code)
    logging.debug("Response Text: %s", response.text)

    try:
        anime_list = json.loads(response.text)
        logging.debug("Anime List: %s", anime_list)
    except json.JSONDecodeError as e:
        logging.error("JSON Decode Error: %s", e)
        return None

    try:
        random_anime = random.choice(anime_list)
        logging.debug("Selected Anime: %s", random_anime)
    except (IndexError, TypeError):
        logging.warning("No anime found in the list using this genre: %s.", genre)
        return None

    name = random_anime['name']
    link = random_anime['link']

    logging.debug("Random Anime: %s", name)
    logging.debug("Link: https://aniworld.to/%s", link)

    return link


def check_internet_connection():
    # return False  # debug
    # offline mini game coming soon!

    try:
        socket.create_connection(("github.com", 80), timeout=5)
        return True
    except OSError:
        pass

    try:
        socket.create_connection(("1.1.1.1", 53), timeout=5)
        return True
    except OSError:
        pass

    try:
        socket.create_connection(("8.8.8.8", 53), timeout=5)
        return True
    except OSError:
        pass

    return False


def get_windows_messagebox_response(message, title, box_type):
    import ctypes  # pylint: disable=import-outside-toplevel
    msg_box_type = {
        "info": 0x40,
        "yesno": 0x04 | 0x20,
        "warning": 0x30,
        "error": 0x10,
    }.get(box_type, 0x40)

    response = ctypes.windll.user32.MessageBoxW(0, message, title, msg_box_type)
    return response == 6 if box_type == "yesno" else True


def get_darwin_messagebox_response(message, title, box_type):
    script = {
        "info": (
            f'display dialog "{message}" with title "{title}" buttons "OK"'
        ),
        "yesno": (
            f'display dialog "{message}" with title "{title}" '
            'buttons {"Yes", "No"}'
        ),
        "warning": (
            f'display dialog "{message}" with title "{title}" '
            'buttons "OK" with icon caution'
        ),
        "error": (
            f'display dialog "{message}" with title "{title}" '
            'buttons "OK" with icon stop'
        ),
    }.get(box_type, (
        f'display dialog "{message}" with title "{title}" buttons "OK"'
    ))

    try:
        result_obj = subprocess.run(
            ["osascript", "-e", script],
            text=True,
            capture_output=True,
            check=False
        )
        return "Yes" in result_obj.stdout if box_type == "yesno" else True
    except subprocess.SubprocessError as e:
        logging.debug("Error showing messagebox on macOS: %s", e)
        return False


def get_linux_messagebox_response(message, title, box_type):
    dialog_program = "zenity" if subprocess.run(
        ["which", "zenity"], capture_output=True, text=True, check=False
    ).returncode == 0 else "kdialog"

    cmd = {
        "zenity": {
            "info": ["zenity", "--info", "--text", message, "--title", title],
            "yesno": ["zenity", "--question", "--text", message, "--title", title],
            "warning": ["zenity", "--warning", "--text", message, "--title", title],
            "error": ["zenity", "--error", "--text", message, "--title", title],
        },
        "kdialog": {
            "info": ["kdialog", "--msgbox", message, "--title", title],
            "yesno": ["kdialog", "--yesno", message, "--title", title],
            "warning": ["kdialog", "--sorry", message, "--title", title],
            "error": ["kdialog", "--error", message, "--title", title],
        }
    }

    cmd = cmd[dialog_program].get(
        box_type,
        [
            "zenity",
            "--info",
            "--text",
            message,
            "--title",
            title
        ]
    )

    try:
        result_obj = subprocess.run(cmd, check=False)
        return (result_obj.returncode == 0) if box_type == "yesno" else True
    except subprocess.SubprocessError as e:
        logging.debug("Error showing messagebox on Linux: %s", e)
        return False


def get_tkinter_messagebox_response(message, title, box_type):
    import tkinter as tk  # pylint: disable=import-outside-toplevel
    from tkinter import messagebox  # pylint: disable=import-outside-toplevel
    root = tk.Tk()
    root.withdraw()

    if box_type == "yesno":
        return messagebox.askyesno(title, message)
    if box_type == "warning":
        messagebox.showwarning(title, message)
    elif box_type == "error":
        messagebox.showerror(title, message)
    else:
        messagebox.showinfo(title, message)
    return True


def show_messagebox(message, title="Message", box_type="info"):
    system = platform.system()

    if system == "Windows":
        return get_windows_messagebox_response(message, title, box_type)

    if system == "Darwin":
        return get_darwin_messagebox_response(message, title, box_type)

    if system == "Linux":
        return get_linux_messagebox_response(message, title, box_type)

    # Fallback for unsupported systems
    return get_tkinter_messagebox_response(message, title, box_type)


def get_current_wallpaper():
    system = platform.system()

    if system == "Windows":
        import ctypes  # pylint: disable=import-outside-toplevel
        buf = ctypes.create_unicode_buffer(512)
        ctypes.windll.user32.SystemParametersInfoW(0x73, len(buf), buf, 0)
        return buf.value

    if system == "Darwin":
        result = os.popen(
            'osascript -e \'tell application "System Events" to '
            'get the picture of the current desktop\''
        ).read().strip()

        if not result:
            result = os.popen(
                'osascript -e \'tell application "System Events" to get the desktop picture\''
            ).read().strip()
        return result

    if system == "Linux":
        try:
            result = os.popen(
                'gsettings get org.gnome.desktop.background picture-uri'
            ).read().strip().strip("'").replace("file://", "")
            return result
        except OSError as e:
            print(f"Could not get current wallpaper: {e}")
            return None

    return None


def set_wallpaper_fit(image_path):
    try:
        import winreg  # pylint: disable=import-error, import-outside-toplevel
        import ctypes  # pylint: disable=import-error, import-outside-toplevel
    except ModuleNotFoundError as e:
        raise ImportError(
            "Required modules (winreg, ctypes) not found. "
            "Ensure you're on Windows."
        ) from e

    key = winreg.OpenKey(
        winreg.HKEY_CURRENT_USER,
        "Control Panel\\Desktop",
        0,
        winreg.KEY_SET_VALUE
    )
    winreg.SetValueEx(key, "WallpaperStyle", 0, winreg.REG_SZ, "6")
    winreg.SetValueEx(key, "TileWallpaper", 0, winreg.REG_SZ, "0")
    winreg.CloseKey(key)

    ctypes.windll.user32.SystemParametersInfoW(20, 0, image_path, 3)


def set_wallpaper(image_path):
    system = platform.system()
    if system == "Windows":
        set_wallpaper_fit(image_path)
    elif system == "Darwin":
        os.system(
            f'osascript -e \'tell application "System Events" to '
            f'set picture of every desktop to "{image_path}"\''
        )
    elif system == "Linux":
        subprocess.call([
            "gsettings",
            "set",
            "org.gnome.desktop.background",
            "picture-uri",
            f"file://{image_path}"
        ])


def minimize_all_windows():
    if platform.system() == "Windows":
        import ctypes  # pylint: disable=import-outside-toplevel
        ctypes.windll.user32.keybd_event(0x5B, 0, 0, 0)  # Press Windows key
        ctypes.windll.user32.keybd_event(0x44, 0, 0, 0)  # Press 'D' key
        ctypes.windll.user32.keybd_event(0x44, 0, 2, 0)  # Release 'D' key
        ctypes.windll.user32.keybd_event(0x5B, 0, 2, 0)  # Release Windows key
    elif platform.system() == "Linux":
        try:
            subprocess.run(["wmctrl", "-k", "on"], check=False)
        except subprocess.SubprocessError as e:
            logging.debug("Error minimizing windows: %s", e)


def show_all_windows():
    if platform.system() == "Linux":
        try:
            subprocess.run(["wmctrl", "-k", "off"], check=True)
        except subprocess.CalledProcessError as e:
            logging.debug("Subprocess error while minimizing windows: %s", e)
        except FileNotFoundError as e:
            logging.debug("wmctrl not found, ensure it's installed: %s", e)


def set_temp_wallpaper():
    # TODO - if no old background set use default windows background
    data = ("aHR0cHM6Ly9naXRodWIuY29tL3Bob2VuaXh0aHJ1c2gvTWFnaWMtRW5naW5l"
            "L2Jsb2IvbWFzdGVyL2xpYnJhcnkvZG9ub3RkZWxldGUucG5nP3Jhdz10cnVl")

    current_wallpaper = get_current_wallpaper()
    logging.debug("Current wallpaper: %s", current_wallpaper)

    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = os.path.join(temp_dir, 'wallpaper.png')

        try:
            response = requests.get(base64.b64decode(data), timeout=15)
            response.raise_for_status()

            with open(file_path, 'wb') as f:
                f.write(response.content)

            set_wallpaper(file_path)
            logging.debug("New wallpaper set: %s", file_path)

            minimize_all_windows()
            if platform.system() == "Darwin":
                show_messagebox(
                    "DO NOT LOOK AT YOUR DESKTOP!\n(DO NOT PRESS FN + F11!!!)",
                    "IMPORTANT!!!",
                    "info"
                )
            elif platform.system() == "Linux":
                time.sleep(5)
                show_all_windows()
            else:
                time.sleep(5)
                minimize_all_windows()

            if current_wallpaper:
                set_wallpaper(current_wallpaper)
                logging.debug("Reverted to original wallpaper: %s", current_wallpaper)
        except requests.RequestException as e:
            logging.debug("Failed to download the wallpaper: %s", e)


def fetch_anime_id(anime_title, season):
    def clean_anime_title(title):
        name = re.sub(r' \(\d+ episodes\)', '', title)
        return re.sub(r'\s+', '%20', name)

    def fetch_mal_data(keyword):
        response = requests.get(
            f"https://myanimelist.net/search/prefix.json?type=anime&keyword={keyword}",
            headers={"User-Agent": aniworld_globals.DEFAULT_USER_AGENT},
            timeout=10
        )
        return response.json() if response.status_code == 200 else None

    def find_best_match(mal_data):
        results = [
            entry for entry in mal_data['categories'][0]['items']
            if 'OVA' not in entry['name']
        ]

        return results[0] if results else None

    def fetch_next_season_id(anime_id):
        url = f"https://myanimelist.net/anime/{anime_id}"
        soup = BeautifulSoup(fetch_url_content(url), 'html.parser')

        sequel_div = soup.find(
            "div",
            string=lambda text: text and "Sequel" in text and "(TV)" in text
        )

        if sequel_div:
            title_div = sequel_div.find_next("div", class_="title")
            link_element = title_div.find("a") if title_div else None

            if link_element:
                match = re.search(r'/anime/(\d+)', link_element.get("href"))
                return match.group(1) if match else None

        return None

    logging.debug("Fetching MAL ID for: %s", anime_title)
    anime_id = None
    keyword = clean_anime_title(anime_title)

    mal_metadata = fetch_mal_data(keyword)
    if not mal_metadata:
        logging.debug("Failed to fetch MyAnimeList data.")
        return None

    best_match = find_best_match(mal_metadata)
    if best_match:
        anime_id = best_match['id']

    while season > 1 and anime_id:
        anime_id = fetch_next_season_id(anime_id)
        if not anime_id:
            logging.debug("Sequel (TV) not found")
            return None
        season -= 1

    return anime_id


def get_description(anime_slug: str):
    url = f"https://aniworld.to/anime/stream/{anime_slug}"

    page_content = fetch_url_content(url)
    soup = BeautifulSoup(page_content, 'html.parser')

    description = soup.find('p', class_='seri_des')['data-full-description']

    return description


def get_description_with_id(anime_title: str, season: int = 1):
    anime_id = fetch_anime_id(anime_title=anime_title, season=season)
    url = f"https://myanimelist.net/anime/{anime_id}"

    page_content = fetch_url_content(url)
    soup = BeautifulSoup(page_content, 'html.parser')
    description = soup.find('meta', property='og:description')['content']
    return description


def install_and_import(package):
    try:
        __import__(package)
    except ImportError:
        while True:
            print(f'The Package "{package}" is not installed!')
            user_input = input(
                f'Do you want me to run “pip install {package}” for you?  (Y|N) '
            ).upper()
            if user_input == "Y":
                print(f"{package} is installing...")
                subprocess.check_call([sys.executable, "-m", "pip", "install", package])
                break
            if user_input == "N":
                sys.exit()
            else:
                clear_screen()
    finally:
        globals()[package] = __import__(package)


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# encoding: utf-8

import argparse
import os
import sys
import re
import logging
import subprocess
import platform
import threading
import random
import signal
import textwrap
from concurrent.futures import ThreadPoolExecutor, as_completed

import npyscreen

from aniworld.search import search_anime
from aniworld import execute, globals as aniworld_globals
from aniworld.common import (
    clear_screen,
    clean_up_leftovers,
    get_season_data,
    set_terminal_size,
    get_version,
    get_language_code,
    is_tail_running,
    get_season_and_episode_numbers,
    setup_anime4k,
    is_version_outdated,
    read_episode_file,
    check_package_installation,
    self_uninstall,
    update_component,
    get_anime_season_title,
    open_terminal_with_command,
    get_random_anime,
    show_messagebox,
    check_internet_connection,
    adventure,
    get_description,
    get_description_with_id
)
from aniworld.extractors import (
    nhentai,
    streamkiste,
    jav,
    hanime
)

from aniworld.globals import DEFAULT_DOWNLOAD_PATH


def format_anime_title(anime_slug):
    logging.debug("Formatting anime title for slug: %s", anime_slug)
    try:
        formatted_title = anime_slug.replace("-", " ").title()
        logging.debug("Formatted title: %s", formatted_title)
        return formatted_title
    except AttributeError:
        logging.debug("AttributeError encountered in format_anime_title")
        sys.exit()


class CustomTheme(npyscreen.ThemeManager):
    default_colors = {
        'DEFAULT': 'WHITE_BLACK',
        'FORMDEFAULT': 'MAGENTA_BLACK',  # Form border
        'NO_EDIT': 'BLUE_BLACK',
        'STANDOUT': 'CYAN_BLACK',
        'CURSOR': 'WHITE_BLACK',  # Text (focused)
        'CURSOR_INVERSE': 'BLACK_WHITE',
        'LABEL': 'CYAN_BLACK',  # Form labels
        'LABELBOLD': 'CYAN_BLACK',  # Form labels (focused)
        'CONTROL': 'GREEN_BLACK',  # Items in form
        'IMPORTANT': 'GREEN_BLACK',
        'SAFE': 'GREEN_BLACK',
        'WARNING': 'YELLOW_BLACK',
        'DANGER': 'RED_BLACK',
        'CRITICAL': 'BLACK_RED',
        'GOOD': 'GREEN_BLACK',
        'GOODHL': 'GREEN_BLACK',
        'VERYGOOD': 'BLACK_GREEN',
        'CAUTION': 'YELLOW_BLACK',
        'CAUTIONHL': 'BLACK_YELLOW',
    }


# pylint: disable=too-many-ancestors, too-many-instance-attributes
class EpisodeForm(npyscreen.ActionForm):
    def create(self):
        logging.debug("Creating EpisodeForm")

        anime_slug = self.parentApp.anime_slug
        logging.debug("Anime slug: %s", anime_slug)

        anime_title = format_anime_title(anime_slug)
        logging.debug("Anime title: %s", anime_title)

        season_data = get_season_data(anime_slug)
        logging.debug("Season data: %s", season_data)

        self.timer = None
        self.start_timer()
        self.setup_signal_handling()

        anime_season_title = get_anime_season_title(slug=anime_slug, season=1)

        def process_url(url):
            logging.debug("Processing URL: %s", url)
            season, episode = get_season_and_episode_numbers(url)
            title = (
                f"{anime_season_title} - Season {season} - Episode {episode}"
                if season > 0
                else f"{anime_season_title} - Movie {episode}"
            )
            return (season, episode, title, url)

        with ThreadPoolExecutor(max_workers=8) as executor:
            future_to_url = {executor.submit(process_url, url): url for url in season_data}

            results = []
            for future in as_completed(future_to_url):
                try:
                    result = future.result(timeout=5)  # Timeout for future result
                    results.append(result)
                    logging.debug("Processed result: %s", result)
                except TimeoutError as e:
                    logging.error("Timeout processing %s: %s", future_to_url[future], e)

        sorted_results = sorted(
            results,
            key=lambda x: (x[0] if x[0] > 0 else 999, x[1])
        )

        season_episode_map = {title: url for _, _, title, url in sorted_results}
        self.episode_map = season_episode_map

        episode_list = list(self.episode_map.keys())
        logging.debug("Episode list: %s", episode_list)

        self.action_selector = self.add(
            npyscreen.TitleSelectOne,
            name="Action",
            values=["Watch", "Download", "Syncplay"],
            max_height=4,
            value=[["Watch", "Download", "Syncplay"].index(aniworld_globals.DEFAULT_ACTION)],
            scroll_exit=True
        )
        logging.debug("Action selector created")

        self.aniskip_selector = self.add(
            npyscreen.TitleSelectOne,
            name="Aniskip",
            values=["Enable", "Disable"],
            max_height=3,
            value=[0 if aniworld_globals.DEFAULT_ANISKIP else 1],
            scroll_exit=True
        )
        logging.debug("Aniskip selector created")

        self.directory_field = self.add(
            npyscreen.TitleFilenameCombo,
            name="Directory:",
            value=aniworld_globals.DEFAULT_DOWNLOAD_PATH
        )
        logging.debug("Directory field created")

        self.language_selector = self.add(
            npyscreen.TitleSelectOne,
            name="Language",
            values=["German Dub", "English Sub", "German Sub"],
            max_height=4,
            value=[
                ["German Dub", "English Sub", "German Sub"].index(
                    aniworld_globals.DEFAULT_LANGUAGE
                )
            ],
            scroll_exit=True
        )
        logging.debug("Language selector created")

        self.provider_selector = self.add(
            npyscreen.TitleSelectOne,
            name="Provider",
            values=[
                "VOE",
                "Vidmoly",
                "Doodstream",
                "SpeedFiles",
                "Vidoza"
            ],
            max_height=6,
            value=[
                [
                    "VOE",
                    "Vidmoly",
                    "Doodstream",
                    "SpeedFiles",
                    "Vidoza"
                ].index(aniworld_globals.DEFAULT_PROVIDER)
            ],
            scroll_exit=True
        )

        logging.debug("Provider selector created")

        self.add(npyscreen.FixedText, value="", editable=False)  # new line
        self.episode_selector = self.add(
            npyscreen.TitleMultiSelect,
            name="Episode Selection",
            values=episode_list,
            max_height=7,
            scroll_exit=True
        )
        logging.debug("Episode selector created")

        self.add(npyscreen.FixedText, value="")

        self.display_text = False

        self.toggle_button = self.add(
            npyscreen.ButtonPress,
            name="Description",
            max_height=1,
            when_pressed_function=self.go_to_second_form,
            scroll_exit=True
        )

        self.action_selector.when_value_edited = self.update_directory_visibility
        logging.debug("Set update_directory_visibility as callback for action_selector")

    def setup_signal_handling(self):
        def signal_handler(_signal_number, _frame):
            try:
                self.parentApp.switchForm(None)
            except AttributeError:
                pass
            self.cancel_timer()
            sys.exit()

        signal.signal(signal.SIGINT, signal_handler)
        logging.debug("Signal handler for SIGINT registered")

    def start_timer(self):
        self.timer = threading.Timer(  # pylint: disable=attribute-defined-outside-init
            random.randint(600, 900),
            self.delayed_message_box
        )
        self.timer.start()

    def cancel_timer(self):
        if self.timer and self.timer.is_alive():
            self.timer.cancel()
            logging.debug("Timer canceled")

    def delayed_message_box(self):
        show_messagebox("Are you still there?", "Uhm...", "info")

    def update_directory_visibility(self):
        logging.debug("Updating directory visibility")
        selected_action = self.action_selector.get_selected_objects()
        logging.debug("Selected action: %s", selected_action)
        if selected_action and selected_action[0] == "Watch" or selected_action[0] == "Syncplay":
            self.directory_field.hidden = True
            self.aniskip_selector.hidden = False
            logging.debug("Directory field hidden, Aniskip selector shown")
        else:
            self.directory_field.hidden = False
            self.aniskip_selector.hidden = True
            logging.debug("Directory field shown, Aniskip selector hidden")
        self.display()

    def on_ok(self):
        logging.debug("OK button pressed")
        self.cancel_timer()
        npyscreen.blank_terminal()
        output_directory = self.directory_field.value if not self.directory_field.hidden else None
        logging.debug("Output directory: %s", output_directory)
        if not output_directory and not self.directory_field.hidden:
            logging.debug("No output directory provided")
            npyscreen.notify_confirm("Please provide a directory.", title="Error")
            return

        selected_episodes = self.episode_selector.get_selected_objects()
        action_selected = self.action_selector.get_selected_objects()
        language_selected = self.language_selector.get_selected_objects()
        provider_selected = self.provider_selector.get_selected_objects()
        aniskip_selected = self.aniskip_selector.get_selected_objects()[0] == "Enable"

        logging.debug("Selected episodes: %s", selected_episodes)
        logging.debug("Action selected: %s", action_selected)
        logging.debug("Language selected: %s", language_selected)
        logging.debug("Provider selected: %s", provider_selected)
        logging.debug("Aniskip selected: %s", aniskip_selected)

        if not (selected_episodes and action_selected and language_selected):
            logging.debug("No episodes or action or language selected")
            npyscreen.notify_confirm("No episodes selected.", title="Selection")
            return

        lang = self.get_language_code(language_selected[0])
        logging.debug("Language code: %s", lang)
        provider_selected = self.validate_provider(provider_selected)
        logging.debug("Validated provider: %s", provider_selected)

        selected_urls = [self.episode_map[episode] for episode in selected_episodes]
        selected_str = "\n".join(selected_episodes)
        logging.debug("Selected URLs: %s", selected_urls)
        npyscreen.notify_confirm(f"Selected episodes:\n{selected_str}", title="Selection")

        if not self.directory_field.hidden:
            output_directory = os.path.join(output_directory)
            os.makedirs(output_directory, exist_ok=True)
            logging.debug("Output directory created: %s", output_directory)

        for episode_url in selected_urls:
            params = {
                'selected_episodes': [episode_url],
                'provider_selected': provider_selected,
                'action_selected': action_selected[0],
                'aniskip_selected': aniskip_selected,
                'lang': lang,
                'output_directory': output_directory,
                'anime_title': format_anime_title(self.parentApp.anime_slug),
                'anime_slug': self.parentApp.anime_slug
            }

            logging.debug("Executing with params: %s", params)
            execute(params)

        if not self.directory_field.hidden:
            logging.debug("Cleaning up leftovers in: %s", output_directory)
            clean_up_leftovers(output_directory)

        self.parentApp.setNextForm(None)
        self.parentApp.switchFormNow()

    def get_language_code(self, language):
        logging.debug("Getting language code for: %s", language)
        return {
            'German Dub': "1",
            'English Sub': "2",
            'German Sub': "3"
        }.get(language, "")

    def validate_provider(self, provider_selected):
        logging.debug("Validating provider: %s", provider_selected)
        valid_providers = ["Vidoza", "Streamtape", "VOE", "Doodstream", "SpeedFiles"]
        while provider_selected[0] not in valid_providers:
            logging.debug("Invalid provider selected, falling back to Vidoza")
            npyscreen.notify_confirm(
                "Doodstream is currently broken.\nFalling back to Vidoza.",
                title="Provider Error"
            )
            self.provider_selector.value = 0
            provider_selected = ["Vidoza"]
        return provider_selected[0]

    def on_cancel(self):
        logging.debug("Cancel button pressed")
        self.cancel_timer()
        self.parentApp.setNextForm(None)

    def go_to_second_form(self):
        self.parentApp.switchForm("SECOND")


# pylint: disable=R0901
class SecondForm(npyscreen.ActionFormV2):
    def create(self):
        anime_slug = self.parentApp.anime_slug
        anime_title = format_anime_title(anime_slug)

        text_content1 = get_description(anime_slug)
        text_content2 = get_description_with_id(anime_title, 1)

        wrapped_text1 = "\n".join(textwrap.wrap(text_content1, width=100))
        wrapped_text2 = "\n".join(textwrap.wrap(text_content2, width=100))

        text_content = f"{wrapped_text1}\n\n{wrapped_text2}"

        self.expandable_text = self.add(
            npyscreen.MultiLineEdit,
            value=text_content,
            max_height=None,
            editable=False
        )

    def on_ok(self):
        self.parentApp.switchForm("MAIN")

    def on_cancel(self):
        self.parentApp.switchForm("MAIN")


class AnimeApp(npyscreen.NPSAppManaged):
    def __init__(self, anime_slug):
        logging.debug("Initializing AnimeApp with slug: %s", anime_slug)
        super().__init__()
        self.anime_slug = anime_slug

    def onStart(self):
        logging.debug("Starting AnimeApp")
        npyscreen.setTheme(CustomTheme)
        version = get_version()
        update_notice = " (Update Available)" if is_version_outdated() else ""
        name = f"AniWorld-Downloader{version}{update_notice}"
        self.addForm(
            "MAIN", EpisodeForm,
            name=name
        )
        self.addForm("SECOND", SecondForm, name="Description")


# pylint: disable=R0912, R0915
def parse_arguments():
    logging.debug("Parsing command line arguments")

    parser = argparse.ArgumentParser(
        description="Parse optional command line arguments."
    )

    # General options
    general_group = parser.add_argument_group('General Options')
    general_group.add_argument(
        '-v', '--version',
        action='store_true',
        help='Print version info'
    )
    general_group.add_argument(
        '-d', '--debug',
        action='store_true',
        help='Enable debug mode'
    )
    general_group.add_argument(
        '-u', '--uninstall',
        action='store_true',
        help='Self uninstall'
    )
    general_group.add_argument(
        '-U', '--update',
        type=str,
        choices=['mpv', 'yt-dlp', 'syncplay', 'all'],
        help='Update mpv, yt-dlp, syncplay, or all.'
    )

    # Search options
    search_group = parser.add_argument_group('Search Options')
    search_group.add_argument(
        '-s', '--slug',
        type=str,
        help='Search query - E.g. demon-slayer-kimetsu-no-yaiba'
    )
    search_group.add_argument(
        '-l', '--link',
        type=str,
        help='Search query - E.g. https://aniworld.to/anime/stream/demon-slayer-kimetsu-no-yaiba'
    )
    search_group.add_argument(
        '-q', '--query',
        type=str,
        help='Search query input - E.g. demon'
    )

    # Episode options
    episode_group = parser.add_argument_group('Episode Options')
    episode_group.add_argument(
        '-e', '--episode',
        type=str,
        nargs='+',
        help='List of episode URLs'
    )
    episode_group.add_argument(
        '-f', '--episode-file',
        type=str,
        help='File path containing a list of episode URLs'
    )
    episode_group.add_argument(
        '-lf', '--episode-local',
        action='store_true',
        help='NOT IMPLEMENTED YET - Use local episode files instead of URLs'
    )

    # Action options
    action_group = parser.add_argument_group('Action Options')
    action_group.add_argument(
        '-a', '--action',
        type=str,
        choices=['Watch', 'Download', 'Syncplay'],
        default=aniworld_globals.DEFAULT_ACTION,
        help='Action to perform'
    )
    action_group.add_argument(
        '-o', '--output',
        type=str,
        help='Download directory E.g. /Users/phoenixthrush/Downloads',
        default=DEFAULT_DOWNLOAD_PATH
    )
    action_group.add_argument(
        '-O', '--output-directory',
        type=str,
        help=(
            'Final download directory, e.g., ExampleDirectory. '
            'Defaults to anime name if not specified.'
        )
    )
    action_group.add_argument(
        '-L', '--language',
        type=str,
        choices=['German Dub', 'English Sub', 'German Sub'],
        default=aniworld_globals.DEFAULT_LANGUAGE,
        help='Language choice'
    )
    action_group.add_argument(
        '-p', '--provider',
        type=str,
        choices=['Vidoza', 'Streamtape', 'VOE',
                 'Doodstream', 'Vidmoly', 'Doodstream', "SpeedFiles"],
        help='Provider choice'
    )

    # Anime4K options
    anime4k_group = parser.add_argument_group('Anime4K Options')
    anime4k_group.add_argument(
        '-A', '--anime4k',
        type=str,
        choices=['High', 'Low', 'Remove'],
        help=(
            'Set Anime4K optimised mode (High, e.g., GTX 1080, RTX 2070, RTX 3060, '
            'RX 590, Vega 56, 5700XT, 6600XT; Low, e.g., GTX 980, GTX 1060, RX 570, '
            'or Remove).'
        )
    )

    # Syncplay options
    syncplay_group = parser.add_argument_group('Syncplay Options')
    syncplay_group.add_argument(
        '-sH', '--syncplay-hostname',
        type=str,
        help='Set syncplay hostname'
    )
    syncplay_group.add_argument(
        '-sU', '--syncplay-username',
        type=str,
        help='Set syncplay username'
    )
    syncplay_group.add_argument(
        '-sR', '--syncplay-room',
        type=str,
        help='Set syncplay room'
    )
    syncplay_group.add_argument(
        '-sP', '--syncplay-password',
        type=str,
        nargs='+',
        help='Set a syncplay room password'
    )

    # Miscellaneous options
    misc_group = parser.add_argument_group('Miscellaneous Options')
    misc_group.add_argument(
        '-k', '--aniskip',
        action='store_true',
        help='Skip intro and outro'
    )
    misc_group.add_argument(
        '-K', '--keep-watching',
        action='store_true',
        help='Continue watching'
    )
    misc_group.add_argument(
        '-r', '--random-anime',
        type=str,
        nargs='?',
        const="all",
        help='Select random anime (default genre is "all", Eg.: Drama)'
    )
    misc_group.add_argument(
        '-D', '--only-direct-link',
        action='store_true',
        help='Output direct link'
    )
    misc_group.add_argument(
        '-C', '--only-command',
        action='store_true',
        help='Output command'
    )
    misc_group.add_argument(
        '-x', '--proxy',
        type=str,
        help='Set HTTP Proxy - E.g. http://0.0.0.0:8080'
    )
    misc_group.add_argument(
        '-w', '--use-playwright',
        action='store_true',
        help='Bypass fetching with a headless browser using Playwright instead (EXPERIMENTAL!!!)'
    )

    args = parser.parse_args()

    if not args.provider:
        if args.action == "Download":
            args.provider = aniworld_globals.DEFAULT_PROVIDER
        else:
            args.provider = aniworld_globals.DEFAULT_PROVIDER_WATCH

    if args.version:
        update_status = " (Update Available)" if is_version_outdated() else ""
        divider = "-------------------" if is_version_outdated() else ""
        banner = fR"""
     ____________________________________{divider}
    < Installed aniworld {get_version()} via {check_package_installation()}{update_status}. >
     ------------------------------------{divider}
            \\   ^__^
             \\  (oo)\\_______
                (__)\\       )\\/\\
                    ||----w |
                    ||     ||
        """

        print(banner)
        sys.exit()

    if args.episode and args.episode_file:
        msg = "Cannot specify both --episode and --episode-file."
        logging.critical(msg)
        print(msg)
        sys.exit()

    if args.debug:
        os.environ['IS_DEBUG_MODE'] = '1'
        aniworld_globals.IS_DEBUG_MODE = True
        logging.getLogger().setLevel(logging.DEBUG)
        logging.debug("============================================")
        logging.debug("Welcome to Aniworld!")
        logging.debug("============================================\n")
        logging.debug("Debug mode enabled")

        if platform.system() == "Darwin":
            if not is_tail_running():
                try:
                    subprocess.run(
                        [
                            "osascript",
                            "-e",
                            'tell application "Terminal" to do script "'
                            'trap exit SIGINT; '
                            'tail -f -n +1 $TMPDIR/aniworld.log" '
                            'activate'
                        ],
                        check=True
                    )
                    logging.debug("Started tailing the log file in a new Terminal window.")
                except subprocess.CalledProcessError as e:
                    logging.error("Failed to start tailing the log file: %s", e)
        elif platform.system() == "Windows":
            try:
                command = ('start cmd /c "powershell -NoExit -c Get-Content '
                           '-Wait \\"$env:TEMP\\aniworld.log\\""')
                subprocess.Popen(command, shell=True)  # pylint: disable=consider-using-with
                logging.debug("Started tailing the log file in a new Terminal window.")
            except subprocess.CalledProcessError as e:
                logging.error("Failed to start tailing the log file: %s", e)
        elif platform.system() == "Linux":
            open_terminal_with_command('tail -f -n +1 /tmp/aniworld.log')

    if args.uninstall:
        self_uninstall()

    if args.update:
        update_component(args.update)
        sys.exit()

    if args.proxy:
        os.environ['HTTP_PROXY'] = args.proxy
        os.environ['HTTPS_PROXY'] = args.proxy
        aniworld_globals.DEFAULT_PROXY = args.proxy
        logging.debug("Proxy set to: %s", args.proxy)

    if args.anime4k:
        setup_anime4k(args.anime4k)

    if args.syncplay_password:
        os.environ['SYNCPLAY_PASSWORD'] = args.syncplay_password[0]
        logging.debug("Syncplay password set.")

    if args.syncplay_hostname:
        os.environ['SYNCPLAY_HOSTNAME'] = args.syncplay_hostname
        logging.debug("Syncplay hostname set.")

    if args.syncplay_username:
        os.environ['SYNCPLAY_USERNAME'] = args.syncplay_username
        logging.debug("Syncplay username set.")

    if args.syncplay_room:
        os.environ['SYNCPLAY_ROOM'] = args.syncplay_room
        logging.debug("Syncplay room set.")

    if args.output_directory:
        os.environ['OUTPUT_DIRECTORY'] = args.output_directory
        logging.debug("Output directory set.")

    if args.use_playwright:
        os.environ['USE_PLAYWRIGHT'] = str(args.use_playwright)
        logging.debug("Playwright set.")

    if not args.slug and args.random_anime:
        args.slug = get_random_anime(args.random_anime)

    return args


def handle_query(args):
    logging.debug("Handling query with args: %s", args)
    if args.query and not args.episode:
        slug = search_anime(query=args.query)
        logging.debug("Found slug: %s", slug)
        season_data = get_season_data(anime_slug=slug)
        logging.debug("Season data: %s", season_data)
        episode_list = list(season_data)
        logging.debug("Episode list: %s", episode_list)

        user_input = input("Please enter the episode (e.g., S1E2): ")
        logging.debug("User input: %s", user_input)
        match = re.match(r"S(\d+)E(\d+)", user_input)
        if match:
            s = int(match.group(1))
            e = int(match.group(2))
            logging.debug("Parsed season: %d, episode: %d", s, e)

        args.episode = [f"https://aniworld.to/anime/stream/{slug}/staffel-{s}/episode-{e}"]
        logging.debug("Set episode URL: %s", args.episode)


def get_anime_title(args):
    logging.debug("Getting anime title from args: %s", args)
    if args.link:
        title = args.link.split('/')[-1]
        logging.debug("Anime title from link: %s", title)
        return title
    if args.slug:
        logging.debug("Anime title from slug: %s", args.slug)
        return args.slug
    if args.episode:
        title = args.episode[0].split('/')[5]
        logging.debug("Anime title from episode URL: %s", title)
        return title
    return None


def main():
    logging.debug("============================================")
    logging.debug("Welcome to Aniworld!")
    logging.debug("============================================\n")
    if not check_internet_connection():
        clear_screen()

        logging.disable(logging.CRITICAL)
        adventure()

        sys.exit()
    try:
        args = parse_arguments()
        logging.debug("Parsed arguments: %s", args)

        validate_link(args)
        handle_query(args)

        language = get_language_code(args.language)
        logging.debug("Language code: %s", language)

        if args.episode_file:
            animes = read_episode_file(args.episode_file)
            for slug, seasons in animes.items():
                if args.output == aniworld_globals.DEFAULT_DOWNLOAD_PATH:
                    args.output = os.path.join(args.output, slug.replace("-", " ").title())
                execute_with_params(args, seasons, slug, language, anime_slug=slug)
            sys.exit()

        anime_title = get_anime_title(args)
        logging.debug("Anime title: %s", anime_title)

        selected_episodes = get_selected_episodes(args, anime_title)

        logging.debug("Selected episodes: %s", selected_episodes)

        if args.episode:
            for episode_url in args.episode:
                slug = episode_url.split('/')[-1]
                execute_with_params(args, selected_episodes, anime_title, language, anime_slug=slug)
            logging.debug("Execution complete. Exiting.")
            sys.exit()
    except KeyboardInterrupt:
        logging.debug("KeyboardInterrupt encountered. Exiting.")
        sys.exit()

    run_app_with_query(args)


def validate_link(args):
    if args.link:
        if args.link.count('/') == 5:
            logging.debug("Provided link format valid.")
        elif args.link.count('/') == 6 and args.link.endswith('/'):
            logging.debug("Provided link format valid.")
            args.link = args.link.rstrip('/')
        else:
            logging.debug("Provided link invalid.")
            args.link = None


def get_selected_episodes(args, anime_title):
    updated_list = None
    if args.keep_watching and args.episode:
        season_data = get_season_data(anime_slug=anime_title)
        logging.debug("Season data: %s", season_data)
        episode_list = list(season_data)
        logging.debug("Episode list: %s", episode_list)

        index = episode_list.index(args.episode[0])
        updated_list = episode_list[index:]
        logging.debug("Updated episode list: %s", updated_list)

    return updated_list if updated_list else args.episode


def check_other_extractors(episode_urls: list):
    logging.debug("Those are all urls: %s", episode_urls)

    jav_urls = []
    nhentai_urls = []
    streamkiste_urls = []
    hanime_urls = []
    remaining_urls = []

    for episode in episode_urls:
        if episode.startswith("https://jav.guru/"):
            jav_urls.append(episode)
        elif episode.startswith("https://nhentai.net/g/"):
            nhentai_urls.append(episode)
        elif episode.startswith("https://streamkiste.tv/movie/"):
            streamkiste_urls.append(episode)
        elif episode.startswith("https://hanime.tv/videos/hentai/"):
            hanime_urls.append(episode)
        else:
            remaining_urls.append(episode)

    logging.debug("Jav URLs: %s", jav_urls)
    logging.debug("Nhentai URLs: %s", nhentai_urls)
    logging.debug("Hanime URLs: %s", hanime_urls)
    logging.debug("Streamkiste URLs: %s", streamkiste_urls)

    for jav_url in jav_urls:
        logging.info("Processing JAV URL: %s", jav_url)
        jav(jav_url)

    for nhentai_url in nhentai_urls:
        logging.info("Processing Nhentai URL: %s", nhentai_url)
        nhentai(nhentai_url)

    for hanime_url in hanime_urls:
        logging.info("Processing hanime URL: %s", hanime_url)
        hanime(hanime_url)

    for streamkiste_url in streamkiste_urls:
        logging.info("Processing Streamkiste URL: %s", streamkiste_url)
        streamkiste(streamkiste_url)

    return remaining_urls


def execute_with_params(args, selected_episodes, anime_title, language, anime_slug):
    selected_episodes = check_other_extractors(selected_episodes)
    logging.debug("Aniworld episodes: %s", selected_episodes)

    params = {
        'selected_episodes': selected_episodes,
        'provider_selected': args.provider,
        'action_selected': args.action,
        'aniskip_selected': args.aniskip,
        'lang': language,
        'output_directory': args.output,
        'anime_title': anime_title.replace('-', ' ').title(),
        'anime_slug': anime_slug,
        'only_direct_link': args.only_direct_link,
        'only_command': args.only_command,
        'debug': args.debug
    }
    logging.debug("Executing with params: %s", params)
    execute(params=params)


def run_app_with_query(args):
    def run_app(query):
        logging.debug("Running app with query: %s", query)
        clear_screen()
        app = AnimeApp(query)
        app.run()

    try:
        try:
            logging.debug("Trying to resize Terminal.")
            set_terminal_size()
            run_app(search_anime(slug=args.slug, link=args.link))
        except npyscreen.wgwidget.NotEnoughSpaceForWidget:
            logging.debug("Not enough space for widget. Asking user to resize terminal.")
            clear_screen()
            print("Please increase your current terminal size.")
            logging.debug("Exiting due to terminal size.")
            sys.exit()
    except KeyboardInterrupt:
        logging.debug("KeyboardInterrupt encountered. Exiting.")
        sys.exit()


if __name__ == "__main__":
    main()



import re


def streamtape_get_direct_link(soup):
    for tag in soup.find_all('script'):
        if "'robotlink')" in tag.text:
            pattern = r"'robotlink'\)\.innerHTML = '(.+?)'\+ \('xcd(.+?)'\)"
            match = re.compile(pattern).search(tag.text)
            if match:
                return f"https:{''.join(match.groups())}"
    return None



import html
import logging
import os
import webbrowser

from typing import List, Dict, Optional
from json import loads
from json.decoder import JSONDecodeError
from urllib.parse import quote

import curses
from bs4 import BeautifulSoup


from aniworld.common import (
    clear_screen,
    fetch_url_content,
    display_ascii_art,
    show_messagebox
)


def search_anime(slug: str = None, link: str = None, query: str = None) -> str:
    clear_screen()
    logging.debug("Starting search_anime function")

    not_found = "Die gewünschte Serie wurde nicht gefunden oder ist im Moment deaktiviert."

    if slug:
        return fetch_by_slug(slug, not_found)

    if link:
        return fetch_by_link(link, not_found)

    return search_by_query(query)


def fetch_by_slug(slug: str, not_found: str) -> str:
    url = f"https://aniworld.to/anime/stream/{slug}"
    logging.debug("Fetching using slug: %s", url)
    response = fetch_url_content(url)
    if response and not_found not in response.decode():
        logging.debug("Found matching slug: %s", slug)
        return slug
    return None


def fetch_by_link(link: str, not_found: str) -> str:
    try:
        logging.debug("Fetching using link: %s", link)
        response = fetch_url_content(link, check=False)
        if response and not_found not in response.decode():
            logging.debug("Found matching slug: %s", link.split('/')[-1])
            return link.split('/')[-1]
    except ValueError:
        logging.debug("ValueError encountered while fetching link")
    return None


def search_by_query(query: str) -> str:
    while True:
        clear_screen()
        if not query:
            print(display_ascii_art())
            query = input("Search for a series: ")
            if query.lower().strip() == "boku no piko":
                show_messagebox("This is not on aniworld...\nThank god...", "Really?...", "info")
        else:
            logging.debug("Using provided query: %s", query)

        url = f"https://aniworld.to/ajax/seriesSearch?keyword={quote(query)}"
        logging.debug("Fetching Anime List with query: %s", query)

        json_data = fetch_anime_json(url)

        if not json_data:
            print("No series found. Try again...")
            query = None
            continue

        if len(json_data) == 1:
            logging.debug("Only one anime found: %s", json_data[0])
            return json_data[0].get('link', 'No Link Found')

        selected_slug = curses.wrapper(display_menu, json_data)
        logging.debug("Found matching slug: %s", selected_slug)
        return selected_slug


def fetch_anime_json(url: str):
    try:
        if os.getenv("USE_PLAYWRIGHT"):
            page_content = html.unescape(fetch_url_content(url).decode('utf-8'))
            soup = BeautifulSoup(page_content, 'html.parser')
            json_data = soup.find('pre').text
        else:
            json_data = html.unescape(fetch_url_content(url).decode('utf-8'))

        decoded_data = loads(json_data)
        if isinstance(decoded_data, list) and decoded_data:
            return decoded_data
    except (AttributeError, JSONDecodeError):
        logging.debug("Error fetching or decoding the anime data.")
    return None


def display_menu(stdscr: curses.window, items: List[Dict[str, Optional[str]]]) -> Optional[str]:
    logging.debug("Starting display_menu function")
    current_row = 0

    konami_code = ['UP', 'UP', 'DOWN', 'DOWN', 'LEFT', 'RIGHT', 'LEFT', 'RIGHT', 'b', 'a']
    entered_keys = []

    key_map = {
        curses.KEY_UP: 'UP',
        curses.KEY_DOWN: 'DOWN',
        curses.KEY_LEFT: 'LEFT',
        curses.KEY_RIGHT: 'RIGHT',
        ord('b'): 'b',
        ord('a'): 'a'
    }

    while True:
        stdscr.clear()
        for idx, anime in enumerate(items):
            name = anime.get('name', 'No Name')
            year = anime.get('productionYear', 'No Year')
            attr = curses.A_REVERSE if idx == current_row else 0
            stdscr.attron(attr)
            stdscr.addstr(idx, 0, f"{name} {year}")
            stdscr.attroff(attr)

        stdscr.refresh()
        key = stdscr.getch()

        if key in key_map:
            entered_keys.append(key_map[key])
            if len(entered_keys) > len(konami_code):
                entered_keys.pop(0)

            if entered_keys == konami_code:
                konami_code_activated()
                entered_keys.clear()
        else:
            entered_keys.clear()

        if key == curses.KEY_DOWN:
            current_row = (current_row + 1) % len(items)
        elif key == curses.KEY_UP:
            current_row = (current_row - 1 + len(items)) % len(items)
        elif key == ord('\n'):
            logging.debug("Selected anime: %s", items[current_row])
            return items[current_row].get('link', 'No Link')
        elif key == ord('q'):
            logging.debug("Exiting menu")
            break

    return None


def konami_code_activated():
    logging.debug("Konami Code activated!")
    curses.endwin()
    webbrowser.open('https://www.youtube.com/watch?v=PDJLvF1dUek')



import os
import shutil
import getpass
import platform
import hashlib
import logging
from typing import Dict, List, Optional, Any

from bs4 import BeautifulSoup


from aniworld.extractors import (
    doodstream_get_direct_link,
    streamtape_get_direct_link,
    vidoza_get_direct_link,
    voe_get_direct_link,
    vidmoly_get_direct_link,
    speedfiles_get_direct_link
)


from aniworld.common import (
    clean_up_leftovers,
    execute_command,
    setup_aniskip,
    fetch_url_content,
    check_dependencies,
    get_language_string,
    get_season_and_episode_numbers,
    print_progress_info,
    countdown,
    sanitize_path,
    setup_autostart,
    setup_autoexit
)


from aniworld.aniskip import aniskip


def providers(soup: BeautifulSoup) -> Dict[str, Dict[int, str]]:
    logging.debug("Extracting provider data from soup")
    try:
        provider_options = soup.find(class_='hosterSiteVideo') \
            .find('ul', class_='row') \
            .find_all('li')

        extracted_data = {}
        for provider in provider_options:
            lang_key = int(provider.get('data-lang-key'))
            redirect_link = provider.get('data-link-target')
            provider_name = provider.find('h4').text.strip()
            if provider_name not in extracted_data:
                extracted_data[provider_name] = {}
            extracted_data[provider_name][lang_key] = f"https://aniworld.to{redirect_link}"
        logging.debug("Extracted provider data: %s", extracted_data)
        return extracted_data
    except AttributeError:
        return None


def build_command(  # pylint: disable=too-many-arguments, too-many-positional-arguments
    link: str, mpv_title: str, player: str, aniskip_selected: bool, selected_provider: str,
    aniskip_options: Optional[List[str]] = None
) -> List[str]:
    logging.debug(
        "Building command for mpv with link: %s, title: %s, player: %s, "
        "aniskip_selected: %s, aniskip_options: %s",
        link, mpv_title, player, aniskip_selected, aniskip_options
    )
    command = [
        player,
        link,
        "--fs",
        "--quiet",
        "--really-quiet",
        "--profile=fast",
        "--hwdec=auto-safe",
        "--video-sync=display-resample",
        f"--force-media-title={mpv_title}"
    ]

    doodstream_referer = "https://dood.li/"
    vidmoly_referer = "https://vidmoly.to/"

    if selected_provider == "Doodstream":
        command.append(f"--http-header-fields=Referer: {doodstream_referer}")
    elif selected_provider == "Vidmoly":
        command.append(f"--http-header-fields=Referer: {vidmoly_referer}")

    if aniskip_selected:
        logging.debug("Aniskip selected, setting up aniskip")
        setup_aniskip()
        if aniskip_options:
            logging.debug("Adding aniskip options: %s", aniskip_options)
            command.extend(aniskip_options)

    logging.debug("Built command: %s", command)
    return command


def build_yt_dlp_command(link: str, output_file: str, selected_provider: str) -> List[str]:
    logging.debug("Building yt-dlp command with link: %s, output_file: %s", link, output_file)
    command = [
        "yt-dlp",
        "--fragment-retries", "infinite",
        "--concurrent-fragments", "4",
        "-o", output_file,
        "--quiet",
        "--no-warnings",
        link,
        "--progress"
    ]

    doodstream_referer = "https://dood.li/"
    vidmoly_referer = "https://vidmoly.to/"

    if selected_provider == "Doodstream":
        command.append("--add-header")
        command.append(f"Referer: {doodstream_referer}")
    elif selected_provider == "Vidmoly":
        command.append("--add-header")
        command.append(f"Referer: {vidmoly_referer}")

    logging.debug("Built yt-dlp command: %s", command)
    return command


def process_aniskip(
    anime_title: str, season_number: int, episode_number: int, anime_slug: str
) -> List[str]:
    logging.debug(
        "Processing aniskip for %s, season %d, episode %d",
        anime_title, season_number, episode_number
    )

    skip_options = aniskip(
        anime_title=anime_title,
        anime_slug=anime_slug,
        episode=episode_number,
        season=season_number
    )
    skip_options_list = skip_options.split(' --')
    processed_options = [
        f"--{opt}" if not opt.startswith('--') else opt
        for opt in skip_options_list
    ]
    logging.debug("Processed aniskip options: %s", processed_options)
    return processed_options


def get_episode_title(soup: BeautifulSoup) -> str:
    logging.debug("Getting episode title from soup")
    german_title_tag = soup.find('span', class_='episodeGermanTitle')
    english_title_tag = soup.find('small', class_='episodeEnglishTitle')

    episode_german_title = german_title_tag.text if german_title_tag else None
    episode_english_title = english_title_tag.text if english_title_tag else None

    episode_title = (
        f"{episode_german_title} / {episode_english_title}"
        if episode_german_title and episode_english_title
        else episode_german_title or episode_english_title
    )

    logging.debug("Episode title: %s", episode_title)
    return episode_title


def get_anime_title(soup: BeautifulSoup) -> str:
    logging.debug("Getting anime title from soup")
    try:
        anime_title = soup.find('div', class_='hostSeriesTitle').text
    except AttributeError:
        logging.warning("Could not use the link provided. Please try using a different one.")
    if 'anime_title' in locals():
        logging.debug("Anime title: %s", anime_title)
        return anime_title
    return None


def get_provider_data(soup: BeautifulSoup) -> Dict[str, Dict[int, str]]:
    logging.debug("Getting provider data from soup")
    data = providers(soup)
    logging.debug("Provider data: %s", data)
    return data


def fetch_direct_link(provider_function, request_url: str) -> str:
    logging.debug("Fetching direct link from URL: %s", request_url)
    html_content = fetch_url_content(request_url)
    soup = BeautifulSoup(html_content, 'html.parser')
    direct_link = provider_function(soup)
    logging.debug("Fetched direct link: %s", direct_link)
    return direct_link


def build_syncplay_command(
    link: str, mpv_title: str, selected_provider: str, aniskip_options: Optional[List[str]] = None
) -> List[str]:
    logging.debug(
        "Building syncplay command with link: %s, title: %s, aniskip_options: %s",
        link, mpv_title, aniskip_options
    )
    syncplay = "SyncplayConsole" if platform.system() == "Windows" else "syncplay"
    anime_title = mpv_title.split(" - ")[0].replace(" ", "_")

    syncplay_password = os.getenv("SYNCPLAY_PASSWORD")
    syncplay_hostname = os.getenv("SYNCPLAY_HOSTNAME")
    syncplay_username = os.getenv("SYNCPLAY_USERNAME")
    syncplay_room = os.getenv("SYNCPLAY_ROOM")

    logging.debug(
        "Syncplay hostname: %s, Syncplay username: %s, Syncplay room: %s",
        syncplay_hostname,
        syncplay_username,
        syncplay_room
    )

    if syncplay_password:
        room_name = (
            f"aniworld-{hashlib.sha256((syncplay_password + anime_title).encode()).hexdigest()}"
        )
    else:
        room_name = f"aniworld-{hashlib.sha256(anime_title.encode()).hexdigest()}"

    if not syncplay_hostname:
        syncplay_hostname = "syncplay.pl:8997"

    if not syncplay_username:
        syncplay_username = getpass.getuser()

    if syncplay_room:
        room_name = syncplay_room

    command = [
        syncplay,
        "--no-gui",
        "--no-store",
        "--host", syncplay_hostname,
        "--name", syncplay_username,
        "--room", room_name,
        "--player-path", shutil.which("mpv"),
    ]
    if syncplay_password:
        password_hash = hashlib.sha256(
            ("aniworld" + syncplay_password + anime_title).encode()
        ).hexdigest()
        command.extend(["--password", password_hash])
    command.extend([
        link,
        "--",
        "--profile=fast",
        "--hwdec=auto-safe",
        "--fs",
        "--video-sync=display-resample",
        f"--force-media-title={mpv_title}"
    ])

    doodstream_referer = "https://dood.li/"
    vidmoly_referer = "https://vidmoly.to/"

    if selected_provider == "Doodstream":
        command.append(f"--http-header-fields=Referer: {doodstream_referer}")
    elif selected_provider == "Vidmoly":
        command.append(f"--http-header-fields=Referer: {vidmoly_referer}")

    if aniskip_options:
        logging.debug("Aniskip options provided, setting up aniskip")
        setup_aniskip()
        command.extend(aniskip_options)

    logging.debug("Built syncplay command: %s", command)
    return command


def perform_action(params: Dict[str, Any]) -> None:
    logging.debug("Performing action with params: %s", params)
    action = params.get("action")
    link = params.get("link")
    mpv_title = params.get("mpv_title")
    anime_title = params.get("anime_title")
    anime_slug = params.get("anime_slug")
    episode_number = params.get("episode_number")
    season_number = params.get("season_number")
    only_command = params.get("only_command", False)
    provider = params.get("provider")
    aniskip_selected = bool(params.get("aniskip_selected", False))

    logging.debug("aniskip_selected: %s", aniskip_selected)

    aniskip_options = process_aniskip_options(
        aniskip_selected=aniskip_selected,
        anime_title=anime_title,
        season_number=season_number,
        episode_number=episode_number,
        anime_slug=anime_slug
    )

    if action == "Watch":
        if not only_command:
            if not platform.system() == "Windows":
                countdown()
        handle_watch_action(
            link, mpv_title, aniskip_selected, aniskip_options, only_command, provider
        )
    elif action == "Download":
        handle_download_action(params)
    elif action == "Syncplay":
        if not only_command:
            if not platform.system() == "Windows":
                countdown()
        setup_autostart()
        setup_autoexit()
        handle_syncplay_action(
            link, mpv_title, aniskip_options, only_command, provider
        )


def process_aniskip_options(
    aniskip_selected: bool,
    anime_title: str,
    season_number: int,
    episode_number: int,
    anime_slug: str
) -> List[str]:
    if aniskip_selected:
        logging.debug("Aniskip is selected, processing aniskip options")
        aniskip_options = process_aniskip(
            anime_title=anime_title,
            season_number=season_number,
            episode_number=episode_number,
            anime_slug=anime_slug
        )
        logging.debug("Aniskip options: %s", aniskip_options)
    else:
        logging.debug("Aniskip is not selected, skipping aniskip options")
        aniskip_options = []
    return aniskip_options


def handle_watch_action(  # pylint: disable=too-many-arguments, too-many-positional-arguments
    link: str,
    mpv_title: str,
    aniskip_selected: bool,
    aniskip_options: List[str],
    only_command: bool,
    selected_provider: str
) -> None:
    logging.debug("Action is Watch")
    mpv_title = mpv_title.replace(" --- ", " - ", 1)
    check_dependencies(["mpv"])
    if not only_command:
        msg = f"Playing '{mpv_title}'"
        if not platform.system() == "Windows":
            print(msg)
        else:
            print_progress_info(msg)
    command = build_command(
        link, mpv_title, "mpv", aniskip_selected, selected_provider, aniskip_options)
    logging.debug("Executing command: %s", command)
    execute_command(command, only_command)
    logging.debug("MPV has finished.\nBye bye!")


def handle_download_action(params: Dict[str, Any]) -> None:
    logging.debug("Action is Download")
    check_dependencies(["yt-dlp"])
    sanitize_anime_title = sanitize_path(params['anime_title'])

    def get_language_from_key(key: int) -> str:
        key_mapping = {
            1: "German Dub",
            2: "English Sub",
            3: "German Sub"
        }

        language = key_mapping.get(key, "Unknown Key")

        if language == "Unknown Key":
            raise ValueError("Key not valid.")

        return language

    output_directory = os.getenv("OUTPUT_DIRECTORY") or params['output_directory']

    file_name = (
        f"{sanitize_anime_title} - S{params['season_number']}E{params['episode_number']}"
        if params['season_number']
        else f"{sanitize_anime_title} - Movie {params['episode_number']}"
    )

    file_path = os.path.join(
        output_directory,
        sanitize_anime_title,
        f"{file_name} ({get_language_from_key(int(params['language']))}).mp4"
    )

    if not params['only_command']:
        msg = f"Downloading to '{file_path}'"
        if not platform.system() == "Windows":
            print(msg)
        else:
            print_progress_info(msg)
    command = build_yt_dlp_command(params['link'], file_path, params['provider'])
    logging.debug("Executing command: %s", command)
    try:
        execute_command(command, params['only_command'])
    except KeyboardInterrupt:
        logging.debug("KeyboardInterrupt encountered, cleaning up leftovers")
        clean_up_leftovers(os.path.dirname(file_path))
    logging.debug("yt-dlp has finished.\nBye bye!")
    if not platform.system() == "Windows":
        print(f"Downloaded to '{file_path}'")
    else:
        print_progress_info(f"Downloaded to '{file_path}'")


def handle_syncplay_action(
    link: str,
    mpv_title: str,
    aniskip_options: List[str],
    only_command: bool,
    selected_provider: str
) -> None:
    logging.debug("Action is Syncplay")
    mpv_title = mpv_title.replace(" --- ", " - ", 1)
    check_dependencies(["mpv", "syncplay"])
    if not only_command:
        msg = f"Playing '{mpv_title}'"
        if not platform.system() == "Windows":
            print(msg)
        else:
            print_progress_info(msg)
    command = build_syncplay_command(link, mpv_title, selected_provider, aniskip_options)
    logging.debug("Executing command: %s", command)
    execute_command(command, only_command)
    logging.debug("Syncplay has finished.\nBye bye!")


def execute(params: Dict[str, Any]) -> None:
    logging.debug("Executing with params: %s", params)
    provider_mapping = {
        "Vidoza": vidoza_get_direct_link,
        "VOE": voe_get_direct_link,
        "Doodstream": doodstream_get_direct_link,
        "Streamtape": streamtape_get_direct_link,
        "Vidmoly": vidmoly_get_direct_link,
        "SpeedFiles": speedfiles_get_direct_link
    }

    selected_episodes = params['selected_episodes']
    action_selected = params['action_selected']
    aniskip_selected = bool(params.get("aniskip_selected", False))
    lang = params['lang']
    output_directory = params['output_directory']
    anime_title = params['anime_title']
    anime_slug = params['anime_slug']
    only_direct_link = params.get('only_direct_link', False)
    only_command = params.get('only_command', False)
    provider_selected = params['provider_selected']

    logging.debug("aniskip_selected: %s", aniskip_selected)

    for episode_url in selected_episodes:
        process_episode({
            'episode_url': episode_url,
            'provider_mapping': provider_mapping,
            'provider_selected': provider_selected,
            'lang': lang,
            'action_selected': action_selected,
            'aniskip_selected': aniskip_selected,
            'output_directory': output_directory,
            'anime_title': anime_title,
            "anime_slug": anime_slug,
            'only_direct_link': only_direct_link,
            'only_command': only_command
        })


def process_episode(params: Dict[str, Any]) -> None:
    try:
        logging.debug("Fetching episode HTML for URL: %s", params['episode_url'])
        episode_html = fetch_url_content(params['episode_url'])
        if episode_html is None:
            logging.debug("No HTML content fetched for URL: %s", params['episode_url'])
            return

        soup = BeautifulSoup(episode_html, 'html.parser')
        episode_title = get_episode_title(soup)
        anime_title = get_anime_title(soup)
        data = get_provider_data(soup)

        logging.debug("Language Code: %s", params['lang'])
        logging.debug("Available Providers: %s", data.keys())

        providers_to_try = [params['provider_selected']] + [
            p for p in params['provider_mapping'] if p != params['provider_selected']
        ]
        for provider in providers_to_try:
            if provider in data:
                process_provider({
                    'provider': provider,
                    'data': data,
                    'lang': params['lang'],
                    'provider_mapping': params['provider_mapping'],
                    'episode_url': params['episode_url'],
                    'action_selected': params['action_selected'],
                    'aniskip_selected': params['aniskip_selected'],
                    'output_directory': params['output_directory'],
                    'anime_title': anime_title,
                    "anime_slug": params['anime_slug'],
                    'episode_title': episode_title,
                    'only_direct_link': params['only_direct_link'],
                    'only_command': params['only_command']
                })
                break
            logging.info("Provider %s not available, trying next provider.", provider)
        else:
            logging.error(
                "Provider %s not found in available providers.",
                params['provider_selected']
            )
    except AttributeError:
        logging.warning("Episode broken.")


def process_provider(params: Dict[str, Any]) -> None:
    logging.debug("Trying provider: %s", params['provider'])
    available_languages = params['data'].get(params['provider'], {}).keys()
    logging.debug("Available Languages for %s: %s", params['provider'], available_languages)

    for language in params['data'][params['provider']]:
        if language == int(params['lang']):
            season_number, episode_number = get_season_and_episode_numbers(params['episode_url'])
            action = params['action_selected']

            provider_function = params['provider_mapping'][params['provider']]
            request_url = params['data'][params['provider']][language]
            link = fetch_direct_link(provider_function, request_url)

            if link is None:
                continue

            if params['only_direct_link']:
                logging.debug("Only direct link requested: %s", link)
                print(link)
                break

            mpv_title = (
                f"{params['anime_title']} --- S{season_number}E{episode_number} - "
                f"{params['episode_title']}"
                if season_number and episode_number
                else f"{params['anime_title']} --- Movie {episode_number} - "
                f"{params['episode_title']}"
            )

            episode_params = {
                "action": action,
                "link": link,
                "mpv_title": mpv_title,
                "anime_title": params['anime_title'],
                "anime_slug": params['anime_slug'],
                "episode_number": episode_number,
                "season_number": season_number,
                "output_directory": params['output_directory'],
                "only_command": params['only_command'],
                "aniskip_selected": params['aniskip_selected'],
                "provider": params['provider'],
                "language": params['lang']
            }

            logging.debug("Performing action with params: %s", episode_params)
            perform_action(episode_params)
            break
    else:
        available_languages = [
            get_language_string(lang_code)
            for lang_code in params['data'][params['provider']].keys()
        ]

        message = (
            f"No available languages for provider {params['provider']} "
            f"matching the selected language {get_language_string(int(params['lang']))}. "
            f"\nAvailable languages: {available_languages}"
        )

        logging.warning(message)
        print(message)



"""
MIT License

Copyright (c) 2024 Phoenixthrush UwU

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import subprocess
import re
import platform
import requests

from bs4 import BeautifulSoup
from aniworld.extractors import voe_get_direct_link
from aniworld import globals as aniworld_globals
from aniworld.common import install_and_import


def clear_screen() -> None:
    if platform.system() == "Windows":
        os.system("cls")
    else:
        os.system("clear")


def fetch_direct_link(link):
    filtered_urls = []

    install_and_import("playwright")
    from playwright.sync_api import sync_playwright  # pylint: disable=import-error, import-outside-toplevel

    with sync_playwright() as p:
        # TODO - add firefox or chromium fallback
        browser = p.webkit.launch(headless=False)
        page = browser.new_page()

        page.on("request", lambda request: filtered_urls.append(request.url)
                if re.match(r"https://voe\.sx/e/.*", request.url) else None)

        page.goto(link)

        # TODO - check for google captcha

        page.wait_for_selector("div.info-right")
        title_element = page.query_selector("div.info-right .title h1")
        title_text = title_element.inner_text().strip() if title_element else "StreamKisteTV"

        page.wait_for_selector("div#single-stream div#stream li.stream div#stream-links a")
        first_stream_link = page.query_selector("div#single-stream div#"
                                                "stream li.stream div#""stream-links a")

        if first_stream_link:
            first_stream_link.click()

        page.wait_for_timeout(5000)
        browser.close()

    try:
        response = requests.get(
            filtered_urls[0],
            timeout=30,
            headers={'User-Agent': aniworld_globals.DEFAULT_USER_AGENT}
        )
    except IndexError:
        # TODO
        # add fallback using another provider or
        # doing the request again with higher timeout and headless=False

        print("No VOE redirect link found.")
        return None

    soup = BeautifulSoup(response.content, 'html.parser')

    # TODO - add other provider options -> Streamtape
    direct_link = voe_get_direct_link(soup)

    return direct_link, title_text


def download_video(link, title):
    filename = os.path.join(os.path.expanduser('~'), 'Downloads', f"{title}.mp4")
    subprocess.run(['yt-dlp', '--quiet', '--no-warnings',
                    '--progress', '-o', filename, link], check=False)


def watch_video(link, title):
    mpv_title = title.replace(" ", "_")
    subprocess.run(['mpv', f'--force-media-title={mpv_title}',
                    '--fs', '--quiet', '--really-quiet', link], check=False)


def streamkiste_get_direct_link(link: str):
    return fetch_direct_link(link)


def streamkiste(episode_link: str = None):
    try:
        if not episode_link:
            clear_screen()
            link = input("Enter the StreamKisteTV link: ")
            action = input("(D)ownload or (W)atch?: ").strip().lower()
        else:
            link = episode_link
            action = "d"

        try:
            link, title = streamkiste_get_direct_link(link)
        except TypeError:
            return

        if action == 'd':
            download_video(link, title)
        elif action == 'w':
            watch_video(link, title)
        else:
            print("Invalid option. Please choose 'D' for download or 'W' for watch.")
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    streamkiste()



from os import environ
from aniworld.common import fetch_url_content


def fetch_ip(url):
    response = fetch_url_content(url)
    return response.decode()


def main():
    url = "http://icanhazip.com"

    print("Without proxy:")
    print(fetch_ip(url))

    proxy = "http://0.0.0.0:8080"

    environ['HTTP_PROXY'] = proxy
    environ['HTTPS_PROXY'] = proxy

    print("With proxy:")
    print(fetch_ip(url), end="")


if __name__ == "__main__":
    main()



"""
MIT License

Copyright (c) 2024 Phoenixthrush UwU

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import platform
import os
import socket


def clear_screen() -> None:
    if platform.system() == "Windows":
        os.system("cls")
    else:
        os.system("clear")


def is_online() -> bool:
    try:
        socket.create_connection(("1.1.1.1", 53), timeout=5)
        return True
    except OSError:
        return False


def display_ascii_art():
    offline = R"""
⣿⣿⣿⣿⠿⢋⣩⣤⣴⣶⣶⣦⣙⣉⣉⣉⣉⣙⡛⢋⣥⣶⣶⣶⣶⣶⣬⡙⢿⣿
⣿⣿⠟⣡⣶⣿⢟⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣦⠙
⣿⢋⣼⣿⠟⣱⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣟⢿⣿⣿⣿⣿⣧
⠃⣾⣯⣿⣾⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣯⣿⣿⡈⢿⣿⣿⣿⣿
⢰⣶⣼⣿⣷⣿⣽⠿⣽⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡌⣿⣷⡀⠛⢿⣿⣿
⢃⣺⣿⣿⣿⢿⠏⢀⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡾⣿⣿⣿⣷⢹⣿⣷⣄⠄⠈⠉
⡼⣻⣿⣷⣿⠏⣰⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣞⣿⣿⣿⠸⣿⣿⣿⣿⣶⣤
⣇⣿⡿⣿⠏⣸⣎⣻⣟⣿⣿⣿⢿⣿⣿⣿⣿⠟⣩⣼⢆⠻⣿⡆⣿⣿⣿⣿⣿⣿
⢸⣿⡿⠋⠈⠉⠄⠉⠻⣽⣿⣿⣯⢿⣿⣿⡻⠋⠉⠄⠈⠑⠊⠃⣿⣿⣿⣿⣿⣿
⣿⣿⠄⠄⣰⠱⠿⠄⠄⢨⣿⣿⣿⣿⣿⣿⡆⢶⠷⠄⠄⢄⠄⠄⣿⣿⣿⣿⣿⣿
⣿⣿⠘⣤⣿⡀⣤⣤⣤⢸⣿⣿⣿⣿⣿⣿⡇⢠⣤⣤⡄⣸⣀⡆⣿⣿⣿⣿⣿⣿
⣿⣿⡀⣿⣿⣷⣌⣉⣡⣾⣿⣿⣿⣿⣿⣿⣿⣌⣛⣋⣴⣿⣿⢣⣿⣿⣿⣿⡟⣿
⢹⣿⢸⣿⣻⣶⣿⢿⣿⣿⣿⢿⣿⣿⣻⣿⣿⣿⡿⣿⣭⡿⠻⢸⣿⣿⣿⣿⡇⢹
⠈⣿⡆⠻⣿⣏⣿⣿⣿⣿⣿⡜⣭⣍⢻⣿⣿⣿⣿⣿⣛⣿⠃⣿⣿⣿⣿⡿⠄⣼
⣦⠘⣿⣄⠊⠛⠿⠿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡟⣼⣿⣿⣿⡿⠁⠄⠟

Aniworld does not work
without internet! :(

To play an offline game next time you're offline,
install Ollama using pip while you're online.

Run these commands:
    pip install ollama
    python -c "import ollama; ollama.pull('llama3.2')"
    """

    return offline


def adventure():
    try:
        import ollama  # pylint: disable=import-error, import-outside-toplevel
    except ModuleNotFoundError:
        print("The 'ollama' module is not installed. Please install it to play the game.")
        return

    try:
        instruction = """
            Use simple English so that non-native speakers can understand what's happening,
            and always ask what "you (the user)" want to do.
            You are the narrator in a text adventure game.
            Start by telling a very short story of about three sentences to set the scene.
            The user will type their answer and you will continue the story based on their input,
            keeping track of what has happened so far.
            Respond by describing what happens next in the story and then ask the user again what they want to do.
            Make sure you are leading the user through an exciting and dynamic adventure.
            Never tell the user that you are an AI!
            You need to describe what's going to happen when the user tells you to look around, for example,
            then tell them what they can see in the current sentence.
            Keep your answers very short, no more than two sentences, and focus on the most important things.
            You need to think about a storyline that does end after some time.
            Also add highlights like monster fights with options as running away or fighting.
            If the player tries todo non logical things like going to a desert while being at the ocean it shouldn't work.
            Also if he tries to use any items that are not logically near you don't allow it,
            you can't equip a sword without buying it somewhere first.
            You can't buy things while fighting.
            Make the fights longer than one prompt, do multiple fight scenes.
        """

        print("Welcome to the Text Adventure Game!")
        print("Type 'exit' at any time to quit the game.\n")

        conversation_history = [{'role': 'system', 'content': instruction}]

        story_start = ollama.chat(
            model='llama3.2',
            messages=conversation_history + [
                {'role': 'user', 'content': 'Begin the adventure with a short story.'}
            ],
            stream=False
        )

        print(f"{story_start['message']['content']}")
        conversation_history.append({
            'role': 'assistant',
            'content': story_start['message']['content']
        })

        while True:
            user_input = input("-> ")

            if user_input.lower() == 'exit':
                print("Thanks for playing! Goodbye!")
                break

            conversation_history.append({'role': 'user', 'content': user_input})

            response = ollama.chat(
                model='llama3.2',
                messages=conversation_history,
                stream=False
            )

            print(f"\n{response['message']['content']}")
            conversation_history.append({
                'role': 'assistant',
                'content': response['message']['content']
            })
    except KeyboardInterrupt:
        return

    # TODO - add exception if model not installed


if __name__ == "__main__":
    adventure()



# pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring

import argparse
import sys
import time
import json
import re

import requests


class SearchResult:
    def __init__(self, slug, title):
        self.title = title
        self.slug = slug

    @property
    def video(self):
        return Video.from_slug(self.slug)

    def __str__(self):
        return f"<Result {self.slug}: {self.title}>"

    __repr__ = __str__


class Video:
    def __init__(self, json_enc):
        self.title = json_enc["hentai_video"]["name"]
        self.slug = json_enc["hentai_video"]["slug"]
        self.sources = {}
        metadata = {}

        for server in json_enc["videos_manifest"]["servers"]:
            for source in server["streams"]:
                if source["url"] != "":
                    name = server["name"]
                    res = source["height"]
                    self.sources[f"{name}-{res}"] = source["url"]

        metadata["brand"] = json_enc["hentai_video"]["brand"]
        metadata["likes"] = json_enc["hentai_video"]["likes"]
        metadata["dislikes"] = json_enc["hentai_video"]["dislikes"]
        metadata["views"] = json_enc["hentai_video"]["views"]
        metadata["tags"] = list(
            map(lambda i: i["text"], json_enc["hentai_video"]["hentai_tags"]))
        metadata["thumbnail"] = json_enc["hentai_video"]["poster_url"]
        metadata["cover"] = json_enc["hentai_video"]["cover_url"]
        metadata["downloads"] = json_enc["hentai_video"]["downloads"]
        metadata["monthly_rank"] = json_enc["hentai_video"]["monthly_rank"]
        metadata["description"] = re.compile(
            r'<[^>]+>').sub("", json_enc["hentai_video"]["description"])
        metadata["franchise_slug"] = json_enc["hentai_franchise"]["slug"]
        metadata["franchise_title"] = json_enc["hentai_franchise"]["title"]
        metadata["franchise_videos"] = [vid["slug"]
                                        for vid in json_enc["hentai_franchise_hentai_videos"]]
        self.metadata = type("Metadata", (), metadata)()

    @staticmethod
    def from_slug(slug):
        r = requests.get(
            f"https://hanime.tv/api/v8/video?id={slug}",
            headers={
                "User-Agent": "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 "
                "(KHTML, like Gecko) Chrome/47.0.2526.106 Safari/537.36"
            },
            timeout=15
        )
        json_enc = r.json()
        return Video(json_enc)

    def at_resolution(self, res):
        max_res = int(max(self.sources, key=lambda source: int(
            source.split("-")[1])).split("-")[1])

        res = min(res, max_res)

        sources = {x: url for x, url in self.sources.items()
                   if x.endswith(str(res))}

        return sources

    def __str__(self):
        return f'<Video {self.slug}: "{self.title}">'

    __repr__ = __str__


def parse_hanime_url(url):
    if "hanime.tv" in url:
        return url.split("/hentai/")[1]

    return None


def download(video, res=1080, verbose=False, folder=False):
    true_res = list(video.at_resolution(res).keys())[0].split("-")[1]
    _ = list(video.at_resolution(res).values())[0]  # source

    if folder:
        out = f"{video.metadata.franchise_slug}/{video.slug}-{true_res}p.mp4"
    else:
        out = f"{video.slug}-{true_res}p.mp4"

    opts = {
        "outtmpl": out
    }

    if not verbose:
        opts["external_downloader_args"] = ["-loglevel", "warning", "-stats"]

    # with YoutubeDL(opts) as dl:
    #    dl.download([source])


def get_random(seed):
    j = requests.get("https://members.hanime.tv/rapi/v7/hentai_videos",
                     params={
                         "source": "randomize",
                         "r": str(seed)
                     },
                     timeout=10
                     ).json()
    results = []

    for result in j["hentai_videos"]:
        results.append(SearchResult(result["slug"], result["name"]))

    return results


# pylint: disable=too-many-arguments, too-many-positional-arguments)
def search(
    query,
    blacklist=None,
    brands=None,
    order_by="title_sortable",
    ordering="asc",
    page=0,
    tags=None,
    tags_mode="AND"
):
    if blacklist is None:
        blacklist = []
    if brands is None:
        brands = []
    if tags is None:
        tags = []

    results = []

    r = requests.post("https://search.htv-services.com/",
                      headers={
                          "User-Agent": "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 "
                          "(KHTML, like Gecko) Chrome/47.0.2526.106 Safari/537.36",
                          "Content-Type": "application/json;charset=UTF-8"
                      },
                      json={
                          "blacklist": blacklist,
                          "brands": brands,
                          "order_by": order_by,
                          "ordering": ordering,
                          "page": page,
                          "search_text": query,
                          "tags": tags,
                          "tags_mode": tags_mode,
                      },
                      timeout=15
                      ).json()

    j = json.loads(r["hits"])

    for result in j:
        results.append(SearchResult(result["slug"], result["name"]))

    return r["nbPages"], results


def roll_search(
    query,
    blacklist=None,
    brands=None,
    order_by="title_sortable",
    ordering="asc",
    _page=0,
    tags=None,
    tags_mode="AND"
):
    if blacklist is None:
        blacklist = []
    if brands is None:
        brands = []
    if tags is None:
        tags = []

    num_pages, results = search(
        query,
        blacklist=blacklist,
        brands=brands,
        order_by=order_by,
        ordering=ordering,
        tags=tags,
        tags_mode=tags_mode
    )

    for p in range(num_pages):
        results += search(
            query,
            blacklist=blacklist,
            brands=brands,
            order_by=order_by,
            ordering=ordering,
            page=p,
            tags=tags,
            tags_mode=tags_mode
        )[1]

    return results


SORT_OPTS_MAP = {
    "upload": "created_at_unix",
    "u": "created_at_unix",
    "views": "views",
    "v": "views",
    "likes": "likes",
    "l": "likes",
    "release": "released_at_unix",
    "r": "released_at_unix",
    "title": "title_sortable",
    "t": "title_sortable"
}

SORT_ORDER_MAP = {
    "a": "asc",
    "ascending": "asc",
    "d": "desc",
    "descending": "desc"
}


def verbose_download(video, res=1080, verbose=False, folder=False):
    print(f"Downloading {video.title}...")
    download(video, res, verbose, folder)


def output(video, args, franchise=False):
    try:
        if args.franchise and franchise:
            if not args.url:
                print(
                    f"Downloading {video.metadata.franchise_title} "
                    "franchise...\n"
                )

            for slug in video.metadata.franchise_videos:
                fran_vid = Video.from_slug(slug)
                output(fran_vid, args, franchise=False)

            return
        if args.url or args.metadata:
            if args.url:
                sources = video.at_resolution(args.resolution)

                print(f"{video.title}:")
                for i, j in sources.items():
                    server, res = tuple(i.split("-"))
                    print(f"{server}, {res}p: {j}")

                print()
            if args.metadata:
                tags_str = ", ".join(video.metadata.tags)
                print(f"URL: https://hanime.tv/videos/hentai/{video.slug}")
                print(f"Brand: {video.metadata.brand}")
                print(f"Franchise: {video.metadata.franchise_title}")
                print(f"Likes: {video.metadata.likes}")
                print(f"Dislikes: {video.metadata.dislikes}")
                print(f"Views: {video.metadata.views}")
                print(f"Downloads: {video.metadata.downloads}")
                print(f"Monthly Rank: {video.metadata.monthly_rank}")
                print(f"Tags: {tags_str}")
                print(f"Description:\n{video.metadata.description}\n")
        else:
            verbose_download(video, args.resolution, args.verbose, args.folder)
    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"Download of {video.title} failed with error \"{e}\"")


# pylint: disable=too-many-locals, too-many-branches, too-many-statements
def hanime(url: str = None):
    if url:
        slug = parse_hanime_url(url)
        video = Video.from_slug(slug)
        if video:
            sources = video.at_resolution(1080)
            print(f"{video.title}:")
            for source_name, source_url in sources.items():
                server, res = source_name.split("-")
                print(f"{server}, {res}p: {source_url}")
            return

        print("Video not found.")
        return

    parser = argparse.ArgumentParser()
    parser.add_argument("video", nargs="*", help="Video URL or search term")
    parser.add_argument("--tags", "-t", help="Tags to search for",
                        action="store", nargs="+", default=[])
    parser.add_argument(
        "--broad-tag-match",
        help="Match videos including any tags specified by --tags",
        action="store_const",
        const="OR",
        default="AND",
    )
    parser.add_argument("--blacklist", "-b", help="Blacklisted tags",
                        action="store", nargs="+", default=[])
    parser.add_argument("--company", "-c", help="Companies/brands to filter by",
                        action="store", nargs="+", default=[])
    parser.add_argument(
        "--page", "-p", help="Page # of search results", default=1, type=int)
    parser.add_argument(
        "--sort-by", "-s",
        help=(
            "Sorting method for search results "
            "([u]pload, [v]iews, [l]ikes, [r]elease, [t]itle)"
        ),
        default="title"
    )
    parser.add_argument(
        "--sort-order", "-w",
        help="Order of sorting ([a]scending or [d]escending)",
        default="ascending",
    )
    parser.add_argument(
        "--roll-search", "-R",
        help=(
            "Roll all search pages into one long page, "
            "useful for large-volume downloads"
        ),
        action="store_true",
        default=False
    )
    parser.add_argument(
        "--resolution", "-r",
        help="Resolution of download, default 1080",
        default=1080,
        type=int,
    )
    parser.add_argument(
        "--index", "-i",
        help="Index of search results to download",
        action="store",
        nargs="+",
        type=int,
        default=[],
    )
    parser.add_argument(
        "--all", "-a",
        help="Download all search results in page",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--folder", "-F",
        help="Create folders by franchise when downloading",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--franchise", "-f",
        help="Download the video and all other videos in its franchise",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--url", "-u",
        help="Show URLs of the source video, do not download",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Show metadata of the source video, do not download",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--verbose", "-v",
        help="Enable verbose logging for video download",
        action="store_true",
        default=False,
    )

    args = parser.parse_args()

    slugs = list(map(parse_hanime_url, args.video))

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    if None not in slugs:
        for slug in slugs:
            video = Video.from_slug(slug)

            output(video, args, args.franchise)
    else:
        query = " ".join(args.video)

        if query == "ALL":
            query = ""

        if query == "random":
            seed = int(time.time() * 1000)
            results = get_random(seed)

            print("Random:")
            if args.index and not args.all:
                for i in args.index:
                    if i <= len(results):
                        output(results[i - 1].video, args, args.franchise)
            else:
                for result in results:
                    if args.all:
                        output(result.video, args, args.franchise)
                    else:
                        print(f"{result.title}")

            sys.exit(0)
        elif query == "new-uploads":
            num_pages, results = search(
                "", order_by="created_at_unix", ordering="desc")
        elif query == "new-releases":
            num_pages, results = search(
                "", order_by="released_at_unix", ordering="desc")
        else:
            sort_by = args.sort_by
            sort_order = args.sort_order

            if sort_by not in SORT_OPTS_MAP:
                print(f'Unknown sort method "{args.sort_by}", using sort by title')
                sort_by = "title"
            if sort_order not in SORT_ORDER_MAP:
                print(f'Unknown sort order "{args.sort_order}", using ascending order')
                sort_order = "ascending"

            search_kwargs = {
                "blacklist": args.blacklist,
                "brands": args.company,
                "tags": args.tags,
                # "page": args.page - 1,
                "tags_mode": args.broad_tag_match,
                "order_by": SORT_OPTS_MAP[sort_by],
                "ordering": SORT_ORDER_MAP[sort_order]
            }

            if args.roll_search:
                num_pages, results = 1, roll_search(query, **search_kwargs)
            else:
                num_pages, results = search(query, **search_kwargs)

        if len(results) > 1 and args.index == [] and not args.all:
            print(f'Found more than one match for "{query}"')
            print(f"Page {args.page} of {num_pages}")
            for index, result in enumerate(results):
                print(f"{index + 1}\t{result.title}")

            print(
                "\nSpecify results to download with --index/-i, "
                "or download all results shown with --all/-a"
            )

        else:
            if len(results) == 0:
                print(f'No results for "{query}"')
            elif args.index and not args.all:
                for i in args.index:
                    if i <= len(results):
                        output(results[i - 1].video, args, args.franchise)
            elif args.all or len(results) == 1:
                for result in results:
                    output(result.video, args, args.franchise)


if __name__ == "__main__":
    hanime()



import subprocess
import re

from aniworld import globals as aniworld_globals
from aniworld.common import clear_screen


class Args:
    def __init__(self, **kwargs):
        self.attributes = {
            'slug': kwargs.get('slug', None),
            'link': kwargs.get('link', None),
            'query': kwargs.get('query', None),
            'episode': kwargs.get('episode', None),
            'action': kwargs.get('action', aniworld_globals.DEFAULT_ACTION),
            'output': kwargs.get('output', aniworld_globals.DEFAULT_DOWNLOAD_PATH),
            'language': kwargs.get('language', aniworld_globals.DEFAULT_LANGUAGE),
            'provider': kwargs.get('provider', aniworld_globals.DEFAULT_PROVIDER),
            'keep_watching': kwargs.get('keep_watching', False),
            'only_direct_link': kwargs.get('only_direct_link', False),
            'only_command': kwargs.get('only_command', False),
            'proxy': kwargs.get('proxy', None),
            'debug': kwargs.get('debug', False)
        }

    def __getattr__(self, item):
        return self.attributes.get(item)

    def __str__(self):
        return f"Args({self.attributes})"


def test_main(args):
    command = ['aniworld']
    arg_map = {
        'slug': '--slug',
        'link': '--link',
        'query': '--query',
        'episode': '--episode',
        'action': '--action',
        'output': '--output',
        'language': '--language',
        'provider': '--provider',
        'proxy': '--proxy',
        'keep_watching': '--keep-watching',
        'only_direct_link': '--only-direct-link',
        'only_command': '--only-command',
        'debug': '--debug'
    }

    for arg, flag in arg_map.items():
        value = getattr(args, arg)
        if value is not None:
            if isinstance(value, bool) and value:
                command.append(flag)
            elif not isinstance(value, bool):
                command.extend([flag, value])

    try:
        result = subprocess.run(
            command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
        return result.stdout
    except subprocess.CalledProcessError as e:
        print(f"Command '{e.cmd}' returned non-zero exit status {e.returncode}.")
        print(f"Error output: {e.stderr}")
        raise


def run_test(test_name, args, expected_pattern, first_test=False):
    try:
        if not first_test:
            print(f"\nTesting {test_name}...")
        else:
            print(f"Testing {test_name}...")
        output = test_main(args)
        assert re.search(expected_pattern, output), (
            f"Output did not contain expected pattern: {output}"
        )
        print("\033[92mOK\033[0m")
    except AssertionError:
        print("\033[91mFAILURE\033[0m")
        print(f"Failed string: {output}")
        print(f"Expected string: {expected_pattern}")


def test_functions():
    clear_screen()
    episode = 'https://aniworld.to/anime/stream/demon-slayer-kimetsu-no-yaiba/staffel-1/episode-1'

    run_test(
        "Download",
        Args(
            only_command=True,
            episode=episode,
            output='/Users/bleh/Downloads/Demon Slayer: Kimetsu no Yaiba - S1E1.mp4',
            language='English Sub',
            provider='Vidoza'
        ),
        r"https://[^\s]*v\.mp4[^\s]*",
        True
    )

    run_test(
        "Watch",
        Args(
            only_command=True,
            episode=episode,
            action='Watch',
            provider='VOE',
            language='German Dub'
        ),
        r"https://[^\s]*master\.m3u8[^\s]*",
        r"mpv 'https://[^\s]*master\.m3u8[^\s]*' --fs --quiet --really-quiet "
        r"'--force-media-title=Demon Slayer: Kimetsu no Yaiba - S1E1 - Grausamkeit / Cruelty'"
    )

    run_test(
        "Syncplay",
        Args(
            only_command=True,
            episode=episode,
            action='Syncplay',
            provider='Vidoza'
        ),
        r"https://[^\s]*v\.mp4[^\s]*",
        r"syncplay --no-gui --no-store --host syncplay.pl:8997 --name phoenixthrush "
        r"--room Demon_Slayer:_Kimetsu_no_Yaiba_-_S1E1_-_Grausamkeit_/_Cruelty "
        r"--player-path /opt/homebrew/bin/mpv 'https://[^\s]*streamtape\.com/get_video[^\s]*' "
        r"-- --fs '--force-media-title=Demon Slayer: Kimetsu no Yaiba - "
        r"S1E1 - Grausamkeit / Cruelty'"
    )


test_functions()



"""
MIT License

Copyright (c) 2024 Phoenixthrush UwU

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import shutil
import subprocess
import sys
import platform
import os
from pathlib import Path


def clear_screen() -> None:
    os.system("cls" if platform.system() == "Windows" else "clear")


def check_for_ytdl():
    if shutil.which("yt-dlp") is None:
        print("yt-dlp not found. Please install it.")
        sys.exit(1)


def check_for_mpv():
    if shutil.which("mpv") is None:
        print("mpv not found. Please install it.")
        sys.exit(1)


def capture_request(route, referrer_data):
    request = route.request
    referrer = request.headers.get("referer")
    if referrer and not referrer_data["referrer"]:
        referrer_data["referrer"] = referrer
    route.continue_()


def capture_response(response, m3u8_counts):
    if response.url.endswith(".m3u8"):
        m3u8_counts[response.url] = m3u8_counts.get(response.url, 0) + 1


def fetch_m3u8_urls(url, browser_type):
    from playwright.sync_api import sync_playwright, Error  # pylint: disable=import-outside-toplevel, import-error
    m3u8_counts = {}
    referrer_data = {"referrer": None}

    with sync_playwright() as p:
        browser = p[browser_type].launch(headless=False)
        page = browser.new_page()

        # Handle popups: Close any popups that open during navigation
        page.on("popup", lambda popup: popup.close())  # Automatically close popups

        page.on("route", lambda route: capture_request(route, referrer_data))
        page.on("response", lambda response: capture_response(response, m3u8_counts))

        page.goto(url)
        page.wait_for_selector("#section-li", timeout=5000)
        page.wait_for_timeout(1000)

        # Click the second li item
        try:
            page.locator('ul#section-li > li').nth(1).locator('a.wp-btn-iframe__shortcode').click()
        except Error as e:
            print(f"Error clicking on the element: {e}")

        page.wait_for_timeout(3000)
        page.wait_for_selector("div.large-screenshot", timeout=3000)
        try:
            for _ in range(3):
                if page.locator("div.large-screenshot").first:
                    page.click("div.large-screenshot")
                    page.wait_for_timeout(2000)
        except Error as e:
            print(f"Error clicking on large-screenshot: {e}")

        page.wait_for_timeout(5000)

        unique_m3u8_urls = [url for url, count in m3u8_counts.items() if count == 1]
        browser.close()
        return unique_m3u8_urls, referrer_data["referrer"]


def process_urls(unique_m3u8_urls, use_mpv, referrer=None):
    print("\nAvailable M3U8 URLs:")
    for idx, url in enumerate(unique_m3u8_urls, start=1):
        print(f"{idx}. {url}")

    choice = int(input("Select the M3U8 URL to download/play (by number): ")) - 1
    selected_url = unique_m3u8_urls[choice]

    referrer_args = ["--referrer", referrer] if referrer else []

    if use_mpv:
        subprocess.run(["mpv", selected_url, *referrer_args], check=False)
    else:
        output_path = os.path.join(os.path.expanduser("~/Downloads"), "%(title)s.%(ext)s")
        subprocess.run(
            ["yt-dlp", selected_url, "-o", output_path, "--no-warnings", *referrer_args],
            check=False
        )


def jav(link: str = None):
    mpv = False
    if not link:
        parser = argparse.ArgumentParser(description="Fetch and download/play video from a URL.")
        parser.add_argument("url", nargs="?", help="URL of the page to fetch.")
        parser.add_argument("--mpv", action="store_true", help="Use mpv instead of yt-dlp.")
        args = parser.parse_args()

        clear_screen()
        input_url = args.url or input("Enter the URL: ")

        if args.mpv:
            check_for_mpv()
            mpv = True
    else:
        input_url = link

    check_for_ytdl()

    try:
        unique_m3u8_urls, referrer = fetch_m3u8_urls(input_url, "webkit")
        if unique_m3u8_urls:
            process_urls(unique_m3u8_urls, mpv, referrer)
        else:
            print("No unique M3U8 URLs found with WebKit. Trying Firefox as a fallback...")

        unique_m3u8_urls, referrer = fetch_m3u8_urls(input_url, "firefox")
        if unique_m3u8_urls:
            process_urls(unique_m3u8_urls, mpv, referrer)
        else:
            print("No unique M3U8 URLs found with Firefox either.")

    except KeyboardInterrupt:
        if not args.mpv:
            for file_extension in [".part", ".ytdl"]:
                for file in Path(".").glob(f"*{file_extension}"):
                    file.unlink()


if __name__ == "__main__":
    jav()



import re
import random
import time

import requests

from aniworld.globals import DEFAULT_USER_AGENT


def doodstream_get_direct_link(soup):
    headers = {
        'User-Agent': DEFAULT_USER_AGENT,
        'Referer': 'https://dood.li/'
    }

    def extract_data(pattern, content):
        match = re.search(pattern, content)
        return match.group(1) if match else None

    def generate_random_string(length=10):
        characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
        return ''.join(random.choice(characters) for _ in range(length))

    # response = requests.get(embeded_doodstream_link, headers=headers)
    # response.raise_for_status()
    content = str(soup)

    pass_md5_pattern = r"\$\.get\('([^']*\/pass_md5\/[^']*)'"
    pass_md5_url = extract_data(pass_md5_pattern, content)
    if not pass_md5_url:
        raise ValueError('pass_md5 URL not found.')

    full_md5_url = f"https://dood.li{pass_md5_url}"

    token_pattern = r"token=([a-zA-Z0-9]+)"
    token = extract_data(token_pattern, content)
    if not token:
        raise ValueError('Token not found.')

    md5_response = requests.get(full_md5_url, headers=headers, timeout=30)
    md5_response.raise_for_status()
    video_base_url = md5_response.text.strip()

    random_string = generate_random_string(10)
    expiry = int(time.time())

    direct_link = f"{video_base_url}{random_string}?token={token}&expiry={expiry}"
    # print(direct_link)

    return direct_link



import re


def vidmoly_get_direct_link(soup):
    scripts = soup.find_all('script')
    file_link_pattern = r'file:\s*"(https?://.*?)"'

    for script in scripts:
        if script.string:
            match = re.search(file_link_pattern, script.string)
            if match:
                file_link = match.group(1)
                return file_link
    return None



import time
import requests
from aniworld.common import fetch_url_content

URL = "https://aniworld.to/anime/stream/demon-slayer-kimetsu-no-yaiba/staffel-1/episode-1"


def measure_fetch_time(fetch_function, url):
    start_time = time.time()
    try:
        content = fetch_function(url)
    except requests.exceptions.RequestException as e:
        return None, f"General error: {str(e)}"
    duration = time.time() - start_time
    return content, duration


content_fetch, fetch_duration = measure_fetch_time(fetch_url_content, URL)
content_requests, requests_duration = measure_fetch_time(
    lambda u: requests.get(u, timeout=5).text,
    URL
)

print(f"Time taken using fetch_url_content: {fetch_duration:.4f} seconds")
print(f"Time taken using requests.get:      {requests_duration:.4f} seconds")



from .provider.doodstream import doodstream_get_direct_link
from .provider.streamtape import streamtape_get_direct_link
from .provider.vidoza import vidoza_get_direct_link
from .provider.voe import voe_get_direct_link
from .provider.vidmoly import vidmoly_get_direct_link
from .provider.speedfiles import speedfiles_get_direct_link

from .nhentai.nhentai import nhentai
from .streamkiste.streamkiste import streamkiste
from .jav.jav import jav
from .hanime.hanime import hanime



from urllib.request import Request, urlopen
from urllib.error import HTTPError, URLError
import logging
import re
from base64 import b64decode

from aniworld import globals as aniworld_globals

REDIRECT_PATTERN = re.compile(r"window\.location\.href\s*=\s*'(https://[^/]+/e/\w+)';")
EXTRACT_VEO_HLS_PATTERN = re.compile(r"'hls': '(?P<hls>.*)'")


def voe_get_direct_link(soup):
    redirect_match = REDIRECT_PATTERN.search(str(soup.prettify))
    if not redirect_match:
        logging.warning("No redirect link found.")
        return None

    redirect_url = redirect_match.group(1)
    try:
        with urlopen(
            Request(
                redirect_url,
                headers={'User-Agent': aniworld_globals.DEFAULT_USER_AGENT}
            ),
            timeout=10
        ) as response:
            redirect_content = response.read()
        redirect_content_str = redirect_content.decode('utf-8')
    except (HTTPError, URLError, TimeoutError) as e:
        logging.error("Failed to fetch URL %s: %s", redirect_url, e)
        return None

    hls_match = EXTRACT_VEO_HLS_PATTERN.search(redirect_content_str)
    if not hls_match:
        logging.warning("No HLS link found.")
        return None

    hls_link = b64decode(hls_match.group("hls")).decode()
    return hls_link



import re
import base64

SPEEDFILES_PATTERN = re.compile(r'var _0x5opu234 = "(?P<encoded_data>.*?)";')


def speedfiles_get_direct_link(soup):
    data = str(soup)
    match = SPEEDFILES_PATTERN.search(data)

    if not match:
        raise ValueError("Pattern not found in the response.")

    encoded_data = match.group("encoded_data")
    decoded = base64.b64decode(encoded_data).decode()
    decoded = decoded.swapcase()[::-1]
    decoded = base64.b64decode(decoded).decode()[::-1]
    decoded_hex = ''.join(chr(int(decoded[i:i + 2], 16)) for i in range(0, len(decoded), 2))
    shifted = ''.join(chr(ord(char) - 3) for char in decoded_hex)
    result = base64.b64decode(shifted.swapcase()[::-1]).decode()

    return result
